////////////////////////////////////////////////////////////////////////////////
/**
 *  球体の自由落下の例
 *
 *  <bullet-src>/Demos/BasicDemo/BasicDemo.cpp を基に作成．
 *  英語コメントはBasicDemo.cppにあったコメント．
 *  日本語コメントは新規に追加したコメント．
 */

// 元のファイル(BasicDemo.cpp)のライセンス表示（ライセンス上必要）
/*
  Bullet Continuous Collision Detection and Physics Library
  Copyright (c) 2003-2006 Erwin Coumans  http://continuousphysics.com/Bullet/

  This software is provided 'as-is', without any express or implied warranty.
  In no event will the authors be held liable for any damages arising from the
  use of this software.
  
  Permission is granted to anyone to use this software for any purpose, 
  including commercial applications, and to alter it and redistribute it freely, 
  subject to the following restrictions:

   1. The origin of this software must not be misrepresented; you must not
   claim that you wrote the original software. If you use this software in
   a product, an acknowledgment in the product documentation would be
   appreciated but is not required.
   
   2. Altered source versions must be plainly marked as such, and must not be
   misrepresented as being the original software.
   
   3. This notice may not be removed or altered from any source distribution.
*/

#include "SimpleTest.h"
#include "GlutStuff.h"
///btBulletDynamicsCommon.h is the main Bullet include file, contains most common include files.
#include "btBulletDynamicsCommon.h"

////////////////////////////////////////////////////////////////////////////////
/**
 *  シミュレーションのメインループから呼ばれるメンバ関数
 */
void SimpleTest::clientMoveAndDisplay()
{
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  // float ms = getDeltaTimeMicroseconds();
  btVector3 pos = m_ball->getCenterOfMassPosition();
  printf("y = %f\n", pos.y());
  
  ///step the simulation
  if (m_dynamicsWorld){
    // m_dynamicsWorld->stepSimulation(ms / 1000000.f);
    m_dynamicsWorld->stepSimulation(0.16);
  }

  // DemoApplication::renderme()はm_dynamicsWorld に登録された物体全てを描画する
  renderme();
  glFlush();
  glutSwapBuffers();
}


////////////////////////////////////////////////////////////////////////////////
/**
 *  'i'キーが押されて，シミュレーション停止状態のときにシーンの描画のために呼ばれる関数
 */
void SimpleTest::displayCallback(void)
{
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  renderme();
  glFlush();
  glutSwapBuffers();

  //btVector3 pc = getCameraPosition();
  //printf("cam pos = %f %f %f\n", pc.getX(), pc.getY(), pc.getZ());
}

////////////////////////////////////////////////////////////////////////////////
/**
 *  シミュレーションの初期化処理
 */
void SimpleTest::initPhysics()
{
  setTexturing(true);
  setShadows(true);
  setCameraDistance(20);
  m_debugMode |= btIDebugDraw::DBG_NoHelpText; // ヘルプ表示を消す

  ///collision configuration contains default setup for memory, collision setup
  m_collisionConfiguration = new btDefaultCollisionConfiguration();

  ///use the default collision dispatcher.
  // For parallel processing you can use a diffent dispatcher (see Extras/BulletMultiThreaded)
  m_dispatcher = new btCollisionDispatcher(m_collisionConfiguration);
  m_broadphase = new btDbvtBroadphase();

  ///the default constraint solver.
  // For parallel processing you can use a different solver (see Extras/BulletMultiThreaded)
  m_solver = new btSequentialImpulseConstraintSolver;
  m_dynamicsWorld = new btDiscreteDynamicsWorld(m_dispatcher, m_broadphase, m_solver, m_collisionConfiguration);

  m_dynamicsWorld->setGravity(btVector3(0, -9.8, 0)); // 重力加速度の設定

  // 地面
  btCollisionShape* groundShape = new btBoxShape(btVector3(50, 50, 50));
  m_collisionShapes.push_back(groundShape); // 終了処理のときにメモリを解放するためポインタを格納．
                                            // 解放が自動で行われないので必要な処理．

  btTransform transform;
  transform.setIdentity();
  transform.setOrigin(btVector3(0, -50, 0)); // 初期位置 (X, Y, Z)

  // DemoApplication::localCreateRigidBody()はbtRigidBodyの生成，
  // そのbtRigidBodyのm_dynamicsWorldへの登録を行う．
  // 指定された形状に応じた慣性モーメント行列の計算も行われる．
  btRigidBody *groundBody = localCreateRigidBody(0, // 質量, 0にすると固定された静止物体として扱われる
						 transform,
						 groundShape);
  groundBody->setRestitution(1.0); // 反発係数．初期値が0なので反発してほしい物体全てに明示的に設定する

  btCollisionShape* colShape = new btSphereShape(1);
  m_collisionShapes.push_back(colShape);

  transform.setIdentity();
  transform.setOrigin(btVector3(0, 10, 0)); // 初期位置 (X, Y, Z)

  m_ball = localCreateRigidBody(1.0, // 質量
                                transform,
                                colShape);
  m_ball->setRestitution(0.8);

  clientResetScene();
}

////////////////////////////////////////////////////////////////////////////////
/**
 *  シミュレーションの終了処理
 */
void SimpleTest::exitPhysics()
{
  //cleanup in the reverse order of creation/initialization

  //remove the rigidbodies from the dynamics world and delete them
  int i;
  for (i=m_dynamicsWorld->getNumCollisionObjects()-1; i>=0 ;i--){
    btCollisionObject* obj = m_dynamicsWorld->getCollisionObjectArray()[i];
    btRigidBody* body = btRigidBody::upcast(obj);
    if (body && body->getMotionState()){
      delete body->getMotionState();
    }
    m_dynamicsWorld->removeCollisionObject( obj );
    delete obj;
  }

  //delete collision shapes
  for (int j=0;j<m_collisionShapes.size();j++){
    btCollisionShape* shape = m_collisionShapes[j];
    delete shape;
  }

  delete m_dynamicsWorld;
  delete m_solver;
  delete m_broadphase;
  delete m_dispatcher;
  delete m_collisionConfiguration;
}
