////////////////////////////////////////////////////////////////////////////////
/**
 *  球体の自由落下のシミュレーション（直接にOpenGLを使う場合）
 *  
 */

#include <btBulletDynamicsCommon.h>
#include <GL/glut.h>
#include <cstdlib>
#include <iostream>

static btDiscreteDynamicsWorld* g_dynamicsWorld; // BPワールド  
static btRigidBody* g_sphereBody; // 球体の力学データ
static btRigidBody* g_groundBody; // 地面の力学データ 
static btSphereShape*      g_sphereShape; // 球体の形状データ
static btStaticPlaneShape* g_groundShape; // 地面の形状データ
static btClock g_clock;

#define XX 0x7f
#define __ 0x40
const int IMG_W = 4, IMG_H = 4;
static unsigned char g_img[] = { // テクスチャ
  XX, XX, __, __, 
  XX, XX, __, __, 
  __, __, XX, XX, 
  __, __, XX, XX, 
};

////////////////////////////////////////////////////////////////////////////////
/** キーボード入力処理 */
void keyboard(unsigned char key, int x, int y){
  if(key == 'q'){
    exit(EXIT_SUCCESS);
  }
}

////////////////////////////////////////////////////////////////////////////////
/** 照明の設定 */
void setLight()
{
  float ambi[] = {0.3, 0.3, 0.3, 1.0}; 
  float diff[] = {1.0, 1.0, 1.0, 1.0}; 
  float spec[] = {0.8, 0.8, 0.8, 1.0}; 
  float pos0[] = {20.0, 20.0, 0.0, 1.0};
  float dir0[] = {0.0, -1.0, 0.0, 0.0}; 

  glLightfv(GL_LIGHT0, GL_AMBIENT,  ambi);
  glLightfv(GL_LIGHT0, GL_DIFFUSE,  diff);
  glLightfv(GL_LIGHT0, GL_SPECULAR, spec);
  
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();{
    glLightfv(GL_LIGHT0, GL_POSITION, pos0);
    glLightfv(GL_LIGHT0, GL_SPOT_DIRECTION, dir0);
  }glPopMatrix();
  glLightf(GL_LIGHT0, GL_SPOT_CUTOFF, 80);

  glEnable(GL_LIGHT0);
  glEnable(GL_LIGHTING);
}

////////////////////////////////////////////////////////////////////////////////
/** 描画 */
void display(){
  float groundColor[] = {0.9f, 0.8f, 0.0f, 1.0f};
  float sphereColor[] = {0.0f, 0.0f, 0.8f, 1.0f};
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  
  setLight();
  //////////
  // 地面を描く
  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, groundColor);
  glEnable(GL_TEXTURE_2D);
  glBegin(GL_QUADS);{
    const double S = 8, T = 8; 
    const double L = 40.0;
    glTexCoord2d(S, T); glVertex3d( L, 0,  L); glNormal3d(0, 1, 0);
    glTexCoord2d(0, T); glVertex3d(-L, 0,  L); glNormal3d(0, 1, 0); 
    glTexCoord2d(0, 0); glVertex3d(-L, 0, -L); glNormal3d(0, 1, 0); 
    glTexCoord2d(S, 0); glVertex3d( L, 0, -L); glNormal3d(0, 1, 0); 
  }glEnd();
  glDisable(GL_TEXTURE_2D);
  
  //////////
  // 球体を描く
  glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, sphereColor);
  glMatrixMode(GL_MODELVIEW);
  glPushMatrix();{
    const int N = 20; // 経緯度方向分割数
    const btVector3 POS = g_sphereBody->getCenterOfMassPosition(); // シミュレーション結果の使用
    glTranslated(POS.x(), POS.y(), POS.z());
    glutSolidSphere(g_sphereShape->getRadius(), N, N);
  }glPopMatrix();
  
  glutSwapBuffers();
}

////////////////////////////////////////////////////////////////////////////////
/** シミュレーション */
void idle(){
  const btScalar USEC_TO_SEC = 1.0/(1000*1000);
  const int MAX_SUBSTEPS = 4;
  const btScalar SUBSTEP_LEN = 0.01; 
  btScalar dt = g_clock.getTimeMicroseconds()*USEC_TO_SEC; // 前フレーム描画からの経過時間の測定
  g_clock.reset(); // 次回の測定の準備   
  g_dynamicsWorld->stepSimulation(dt, MAX_SUBSTEPS, SUBSTEP_LEN); // 実経過時間に合わせてシミュレーションを進める
  glutPostRedisplay();
}

////////////////////////////////////////////////////////////////////////////////
/** カメラ設定 */
void reshape(int w, int h){
  const double FOV = 60.0;
  const double NEAR_CLIP = 1.0;
  const double FAR_CLIP  = 400.0;
  double aspect = ((double)w)/h;
  glViewport(0, 0, w, h);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(FOV, aspect, NEAR_CLIP, FAR_CLIP);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  gluLookAt(0, 25, -35, // カメラ視点位置 (X, Y, Z)
            0, 0, 0,   // 注視点 (X, Y, Z)
            0, 1, 0);  // カメラの「上」方向 (X, Y, Z)
}

////////////////////////////////////////////////////////////////////////////////
/** GLUT使用準備 */
void setUpGLUT(int argc, char *argv[]){
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);
  glutInitWindowSize(640, 480);
  glutInitWindowPosition(0, 0);
  glutCreateWindow("BP First Step");

  glutDisplayFunc(display);
  glutIdleFunc(idle);
  glutKeyboardFunc(keyboard);
  glutReshapeFunc(reshape);
}

////////////////////////////////////////////////////////////////////////////////
/** OpenGL使用準備 */
void setUpGL(){
  glTexImage2D(GL_TEXTURE_2D,
               0, // mipmap処理レベル
               GL_LUMINANCE_ALPHA, IMG_W, IMG_H,
               0, // 縁なし
               GL_LUMINANCE, GL_UNSIGNED_BYTE, g_img);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
  glTexParameterf(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
  
  glEnable(GL_DEPTH_TEST);
  glClearColor(0.0, 0.0, 0.0, 1.0);
}

////////////////////////////////////////////////////////////////////////////////
/** BP使用準備 */
void setUpBP(){
  btDefaultCollisionConfiguration* collisionConfiguration = new btDefaultCollisionConfiguration();
  btCollisionDispatcher* dispatcher = new btCollisionDispatcher(collisionConfiguration);
  btBroadphaseInterface* broadphase = new btDbvtBroadphase();
  btConstraintSolver*    solver     = new btSequentialImpulseConstraintSolver;
  g_dynamicsWorld = new btDiscreteDynamicsWorld(dispatcher, broadphase, solver,
                                                collisionConfiguration);

  btContactSolverInfo& slvInfo = g_dynamicsWorld->getSolverInfo();
  slvInfo.m_erp  = 0.0;
  
  g_dynamicsWorld->setGravity(btVector3(0, -9.8, 0)); // 重力加速度の設定
}


////////////////////////////////////////////////////////////////////////////////
/** BP使用終了処理 */
void finishBP(){
  g_dynamicsWorld->removeRigidBody(g_sphereBody);
  delete g_sphereBody->getMotionState();
  delete g_sphereBody;
  g_dynamicsWorld->removeRigidBody(g_groundBody);
  delete g_groundBody->getMotionState();
  delete g_groundBody;
  delete g_sphereShape;
  delete g_groundShape;
  
  delete g_dynamicsWorld->getConstraintSolver();
  delete g_dynamicsWorld->getBroadphase();
  btCollisionDispatcher* dispatcher = (btCollisionDispatcher*)g_dynamicsWorld->getDispatcher();
  delete dispatcher->getCollisionConfiguration(); 
  delete g_dynamicsWorld->getDispatcher();
  delete g_dynamicsWorld;
}

////////////////////////////////////////////////////////////////////////////////
/**
 *  物体を作成して，ワールドに配置する
 */
void setUpObjects(){
  btTransform transform; // 座標変換データ
  transform.setIdentity(); // 座標変換データの初期化（回転なし，平行移動なし）

  //////////
  // 地面の作成（ワールド座標系の Z-X 平面に一致させる）
  g_groundShape = new btStaticPlaneShape(btVector3(0, 1, 0), // 平面の法線ベクトル
                                         0); // 原点から平面への最短距離
  btDefaultMotionState* groundMS = new btDefaultMotionState(transform);
  btRigidBody::btRigidBodyConstructionInfo
    groundCI(0,                   // 質量（0 -> 固定） 
             groundMS,            // 初期位置，姿勢
             g_groundShape,       // 形状データ
             btVector3(0, 0, 0)); // ローカル座標系各軸まわりの慣性モーメント(Ix, Iy, Iz)
  
  g_groundBody = new btRigidBody(groundCI);
  g_groundBody->setRestitution(1.0); // 反発パラメータの設定
  g_dynamicsWorld->addRigidBody(g_groundBody);
  
  //////////
  // 球体の作成
  g_sphereShape = new btSphereShape(1); // 形状データの作成
  transform.setOrigin(btVector3(0, 10, 0)); // (X, Y, Z)の平行移動に設定する
  btDefaultMotionState* sphereMS = new btDefaultMotionState(transform);
  btVector3 sphereInertia;
  const btScalar SPHERE_MASS = 1.0;
  g_sphereShape->calculateLocalInertia(SPHERE_MASS, sphereInertia);  
  btRigidBody::btRigidBodyConstructionInfo
    sphereCI(SPHERE_MASS,    // 質量
             sphereMS,       // 初期位置，姿勢
             g_sphereShape,  // 形状データ
             sphereInertia); // ローカル座標系各軸まわりの慣性モーメント(Ix, Iy, Iz)

  g_sphereBody = new btRigidBody(sphereCI);
  g_sphereBody->setRestitution(1.0); // 反発パラメータの設定
  g_dynamicsWorld->addRigidBody(g_sphereBody);
}

////////////////////////////////////////////////////////////////////////////////
int main(int argc, char *argv[]){
  setUpGLUT(argc, argv); // ウィンドウを準備する
  setUpGL();      // 3次元描画の準備を行う
  setUpBP();      // BP使用準備を行う
  setUpObjects(); // 物体を作成して配置する
  
  atexit(finishBP); // 強制終了時の対策
  glutMainLoop();   // イベント処理ループ
  finishBP();       // BPの後始末を行う
  return 0;
}
