////////////////////////////////////////////////////////////////////////////////
/**
 *  投げ上げ運動のシミュレーション
 *  
 */

#include "BPWorld.h" // BPWorldを使うため
#include <cstdio> 

btRigidBody *g_sphereBody; //* 球体の力学データ
btRigidBody *g_targetBody; //* 的の力学データ
int g_angle = 0; //* 射出仰角 [deg]
bool g_shotRequested = false; //* 射出するか否かを指定するフラグ
BPWorld *g_bpWorld;

////////////////////////////////////////////////////////////////////////////////
/**
 *  シミュレーションのステップごとに呼ばれる関数（コールバック）
 *  物体の状態をモニタリングするために使う．
 */
void doMonitoring(btScalar currentTime){
  const btScalar V = 100; // 初速度の大きさ
  
  if(g_shotRequested){
    // この物体についてのシミュレーションを有効にする．
    // 一定時間静止していた物体は休眠（sleep）状態になっており
    // そのままsetLinearVelocity()を呼び出しても効果がない
    g_sphereBody->activate(true);

    btScalar ang = btRadians(g_angle);
    
    // 物体に速度を与える
    g_sphereBody->setLinearVelocity
      (btVector3(0, V*btSin(ang), V*btCos(ang))); // 速度ベクトル
    g_shotRequested = false;
  }
  
  // 物体を追従するようにカメラの位置，姿勢を指定する
  btVector3 pc = g_sphereBody->getCenterOfMassPosition();
  g_bpWorld->setCamera(pc.x(), pc.y(), pc.z(), // 注視点の位置 [x, y, z]
                       20, // 注視点からカメラ視点までの距離
                       95, // 方位角（azimuth） [deg]
                       13.2, // 仰角（elevation） [deg]
                       0, 1, 0); // 天井方向ベクトル [x, y, z] （cf. OpenGLのgluLookAt()）
}


////////////////////////////////////////////////////////////////////////////////
/**
 *  キー入力処理
 */
void checkKey(int key, int x, int y)
{
  const int DELTA = 1; // [deg]
  const int LL_ANGLE = -90; // 角度下限 [deg]
  const int UL_ANGLE =  90; // 角度上限 [deg]
  
  switch(key){
  case 'j':
    g_angle += DELTA;
    printf("angle = %d deg\n", g_angle);
    break;
    
  case 'k':
    g_angle -= DELTA;
    printf("angle = %d deg\n", g_angle);
    break;
    
  case '0':
    g_shotRequested = true;
    puts("Shoot\n");
    break;
  }

  if(g_angle > UL_ANGLE){
    g_angle = UL_ANGLE;
  }
  if(g_angle < LL_ANGLE){
    g_angle = LL_ANGLE;
  }
}

////////////////////////////////////////////////////////////////////////////////
/**
 *  物体を作成して，ワールドに配置する
 */
void setUpObjects(BPWorld& bpWorld){
  btSphereShape *sphereShape = new btSphereShape(1); // 球体の形状データ
  btSphereShape *targetShape = new btSphereShape(1); // 的の形状データ
  
  btTransform transform; // 座標変換データ
  transform.setIdentity(); // 座標変換データの初期化（回転なし，平行移動なし）
  transform.setOrigin(btVector3(0, sphereShape->getRadius(), -510)); // (X, Y, Z)の平行移動に設定する

  // 球体の作成
  g_sphereBody = bpWorld.createRigidBody(1.0,       // 質量
                                         transform, // 初期位置，姿勢
                                         sphereShape);    // 形状データ

  g_sphereBody->setRestitution(1.0); // 反発パラメータの設定
  g_sphereBody->setFriction(0.0); // 摩擦パラメータの設定

  transform.setOrigin(btVector3(0, targetShape->getRadius(), 0)); // 的の位置
  
  // 的の作成
  g_targetBody = bpWorld.createRigidBody(1.0,
                                         transform,
                                         targetShape);
  g_targetBody->setRestitution(1.0); // 反発パラメータの設定
  g_targetBody->setFriction(0.0); // 摩擦パラメータの設定
}


////////////////////////////////////////////////////////////////////////////////
/**
 *  使い方を表示する
 */
void showUsage()
{
  printf("--- Usage ----------------------------------------\n"
         " Press 'j'/'k' to increase/decrease the shooting angle\n"
         " Press '0' (zero) to shoot the sphere\n"
         " Press Space to reset the scene\n"
         "--------------------------------------------------\n");
}

////////////////////////////////////////////////////////////////////////////////
/**
 *  プログラムの開始点
 */
int main(){
  // 動力学シミュレーションのワールドの準備
  BPWorld bpWorld("Ball Shooting", // ウィンドウのタイトル
                  640, 480,        // ウィンドウの幅と高さ [pixels]
                  doMonitoring,    // モニタリング用関数へのポインタ
                  checkKey);       // キーボード入力処理用関数へのポインタ

  g_bpWorld = &bpWorld; // 物体を追従するようにカメラを動かす処理の準備
  // bpWorld.enableFixedTimeStep(true); // 固定ステップを有効にする場合
  // bpWorld.setTimeStep(0.001);
  bpWorld.setGlobalERP(0.0);
  
  setUpObjects(bpWorld); // 物体を作成して配置する
  showUsage();
  bpWorld.start(); // ウィンドウを表示して，シミュレーションを開始する

  return 0;
}
