////////////////////////////////////////////////////////////////////////////////
/**
 *  ビリヤードのシミュレーション（的球が1個の場合）
 *
 */

#include "BPWorld.h" // BPWorldを使うため
#include <cstdio> 

btRigidBody *g_cueBallBody; // 手球の力学データ
int g_angle = 0; // 射出角 [deg]
bool g_shotRequested = false;

////////////////////////////////////////////////////////////////////////////////
/**
 *  シミュレーションのステップごとに呼ばれる関数（コールバック）
 *  物体の状態をモニタリングするために使う．
 */
void doMonitoring(btScalar currentTime){
  const btScalar F = 20; // 力の大きさ
  
  if(g_shotRequested){
    // この物体についてのシミュレーションを有効にする．
    // 一定時間静止していた物体は休眠（sleep）状態になっており
    // そのままapplyCentralImpulse()を呼び出しても効果がない
    g_cueBallBody->activate(true);

    btScalar ang = btRadians(g_angle);
    
    // 物体に撃力（瞬間的にかかる力）を加える．作用点は球体の中心．
    g_cueBallBody->applyCentralImpulse
      (btVector3(F*btSin(ang), 0, F*btCos(ang))); // 撃力ベクトル
    g_shotRequested = false;
  }
}


////////////////////////////////////////////////////////////////////////////////
/**
 *  キー入力処理
 */
void checkKey(int key, int x, int y)
{
  const int DELTA = 1; // [deg]
  const int LL_ANGLE = -90; // 角度下限 [deg]
  const int UL_ANGLE =  90; // 角度上限 [deg]
  
  switch(key){
  case 'j':
    g_angle += DELTA;
    printf("angle = %d deg\n", g_angle);
    break;
    
  case 'k':
    g_angle -= DELTA;
    printf("angle = %d deg\n", g_angle);
    break;
    
  case '0':
    g_shotRequested = true;
    puts("Shoot\n");
    break;
  }

  if(g_angle > UL_ANGLE){
    g_angle = UL_ANGLE;
  }
  if(g_angle < LL_ANGLE){
    g_angle = LL_ANGLE;
  }
}

////////////////////////////////////////////////////////////////////////////////
/**
 *  物体を作成して，ワールドに配置する
 */
void setUpObjects(BPWorld& bpWorld){
  const btScalar R = 1;
  btSphereShape *ballShape = new btSphereShape(R); // 球体の形状データ
  btTransform transform; // 座標変換データ  
  transform.setIdentity(); // 座標変換データの初期化

  transform.setOrigin(btVector3(0, R, -15));

  // 手球
  g_cueBallBody = bpWorld.createRigidBody(1.0,        // 質量
                                          transform,  // 初期位置，姿勢
                                          ballShape); // 形状データ
  g_cueBallBody->setRestitution(1.0); // 反発パラメータの設定
  g_cueBallBody->setFriction(0.0);  // 摩擦パラメータの設定


  transform.setOrigin(btVector3(0, R, 0));

  // 的球．複数の物体で同一の形状データを共用してよい
  btRigidBody *objBallBody = bpWorld.createRigidBody(1.0,        // 質量
                                                     transform,  // 初期位置，姿勢
                                                     ballShape); // 形状データ
  objBallBody->setRestitution(1.0); // 反発パラメータの設定
  objBallBody->setFriction(0.5);  // 摩擦パラメータの設定
  objBallBody->setAngularFactor(btVector3(0, 0, 0)); // 回転を無効にする
}

////////////////////////////////////////////////////////////////////////////////
/**
 *  使い方を表示する
 */
void showUsage()
{
  printf("--- Usage ----------------------------------------\n"
         " Press 'j'/'k' to increase/decrease the shooting angle\n"
         " Press '0' (zero) to shoot the sphere\n"
         " Press Space to reset the scene\n"
         "--------------------------------------------------\n");
}

////////////////////////////////////////////////////////////////////////////////
/**
 *  プログラムの開始点
 */
int main(){
  // 動力学シミュレーションのワールドの準備
  BPWorld bpWorld("Billiards - One ball",  // ウィンドウのタイトル
                  640, 480,     // ウィンドウの幅と高さ [pixels]
                  doMonitoring, // モニタリング用関数へのポインタ
                  checkKey);    // キーボード入力処理用関数へのポインタ

  bpWorld.setGlobalERP(0.0);
  
  setUpObjects(bpWorld); // 物体を作成して配置する
  showUsage();
  bpWorld.start(); // ウィンドウを表示して，シミュレーションを開始する

  return 0;
}
