////////////////////////////////////////////////////////////////////////////////
/**
 *  断面が円形の物体の転がり運動のシミュレーション
 *
 */

#include "BPWorld.h" // BPWorldを使うため
#include <iostream>

////////////////////////////////////////////////////////////////////////////////
/**
 *  ユーザに数値を問い合わせる
 */
btScalar queryValue(const char *msg)
{
  using namespace std;
  btScalar val;
  cout << msg << flush;
  cin >> val;
  
  return val;
}

////////////////////////////////////////////////////////////////////////////////
/**
 *  物体を作成して，ワールドに配置する 
 */
void setUpObjects(BPWorld& bpWorld){
  const btScalar R1 = 0.5; // 物体1の断面の半径
  const btScalar R2 = 0.5; // 物体2の断面の半径
  btCollisionShape *shape1 = new btSphereShape(R1); // 物体1の形状
  btCollisionShape *shape2 = new btCylinderShapeZ(btVector3(R2, R2, 0.5)); // 物体2の形状
  // btCollisionShape *shape2 = new btSphereShape(R2); // 物体2の形状
  const btScalar D = 0.2; // 板の厚さ
  const btScalar W = 12;   // 板の幅
  const btScalar Z_OFFSET = -4;
  btScalar l;   // 板の長さ
  const btScalar L_CAP = 0.4; // 球の接点から板の上端までの長さ
  btScalar th; // 板の傾斜角度 [rad]
  btScalar thDeg;  // 板の傾斜角度 [deg]
  btScalar h; // 重心の高さ
  const btScalar XC = 12; // 板先端の地面接触点のx座標
  btTransform trans1; // 座標変換データ
  btTransform trans2;
  btTransform trans3;
  btTransform transAll;
  btQuaternion quat;

  thDeg = queryValue("slope angle [deg] = ");
  th = btRadians(thDeg);
  l = XC/btCos(th) + (D + R1)*btTan(th) + L_CAP;
  h = XC*btTan(th) + (D + R1)/btCos(th); // 板に接するように設定する
  
  trans1.setIdentity(); 
  trans1.setOrigin(btVector3(0, h, 0)); 

  btRigidBody *body1 = bpWorld.createRigidBody(1.0,    // 質量
                                               trans1, // 初期位置，姿勢
                                               shape1);   // 形状データ
  
  // XY平面でみて，物体1と板の接点 = 物体2と板の接点 となるように配置する
  trans1.setIdentity(); 
  trans1.setOrigin(btVector3((R2 - R1)*btSin(th),
                             (XC + R1*btSin(th))*btTan(th) + D/btCos(th) + R2*btCos(th),
                             Z_OFFSET)); 
  btRigidBody *body2 = bpWorld.createRigidBody(1.0,    // 質量
                                               trans1, // 初期位置，姿勢
                                               shape2);    // 形状データ
  
  // 板のの形状データ
  btCollisionShape* boardShape = new btBoxShape(btVector3(0.5*l, 0.5*D, 0.5*W));

  // 板の配置: 球に接するように配置する（前提: 球の中心がY軸上にある）
  trans1.setIdentity();
  trans1.setOrigin(btVector3(-0.5*l, 0.5*D, 0)); // 板の角を原点に置く
  
  trans2.setIdentity();
  quat.setEuler(0, 0, -th); // Z軸まわりに板を回転
  trans2.setRotation(quat);

  trans3.setIdentity();
  trans3.setOrigin(btVector3(XC, 0, 0)); // X軸方向に平行移動

  transAll.mult(trans2, trans1);
  transAll.mult(trans3, transAll); // 座標変換実行時にはtrans1, 2, 3の順に適用される
  
  btRigidBody *boardBody = bpWorld.createRigidBody(0, // 質量 0 -> 固定
                                                   transAll,
                                                   boardShape);

  // 摩擦係数の設定．摩擦係数 mu =  lambda1 * lambda2,
  // lambda1, lambda2はsetFriction()で接触する2物体それぞれに指定する値
  //  各物体について回転に必要な最小値より大きな値を設定する
  body1->setFriction((2.0/7)*tan(th)); // 球体の場合
  body2->setFriction((1.0/3)*tan(th)); // 円筒の場合
  boardBody->setFriction(1.0);
}

////////////////////////////////////////////////////////////////////////////////
/**
 *  プログラムの開始点
 */
int main(){
  // 動力学シミュレーションのワールドの準備
  BPWorld bpWorld("Ball Rolling Down", // ウィンドウのタイトル
                  640, 480);         // ウィンドウの幅と高さ [pixels]
  setUpObjects(bpWorld); // 物体を作成して配置する

  bpWorld.toggleIdle(); // 一時停止状態から始める
  bpWorld.start(); // ウィンドウを表示して，シミュレーションを開始する
  
  return 0;
}
