////////////////////////////////////////////////////////////////////////////////
/**
 *  ドアのシミュレーション
 *
 */

#include "BPWorld.h" // BPWorldを使うため
#include <cstdio> // C言語のstdio.hを使うときのC++式の書き方 


btHingeConstraint *g_joint; // ヒンジ ジョイント

////////////////////////////////////////////////////////////////////////////////
/**
 *  シミュレーションのステップごとに呼ばれる関数（コールバック）
 *  物体の状態をモニタリングするために使う．
 */
void doMonitoring(btScalar currentTime){
  printf("angle = %7.3f\n", btDegrees(g_joint->getHingeAngle()));
}

////////////////////////////////////////////////////////////////////////////////
/**
 *  物体を作成して，ワールドに配置する 
 */
void setUpObjects(BPWorld& bpWorld){
  const btScalar H = 6; // ドアの高さ
  const btScalar W = 2; // ドアの幅
  const btScalar D = 0.2; // ドアの厚さ
  const btScalar OFFSET = 0.1; // 地面との隙間
  // ドアの形状データ
  btCollisionShape* doorShape = new btBoxShape(btVector3(0.5*W, 0.5*H, 0.5*D));
  // ドアを固定する柱（door jamb）の形状データ
  btCollisionShape* jambShape = new btBoxShape(btVector3(0.5*D, 0.5*H, 0.5*D)); 
  
  btTransform transform;
  transform.setIdentity();
  transform.setOrigin(btVector3(-0.5*W, 0.5*H + OFFSET, -0.5*D)); // 初期位置 (X, Y, Z)
  btRigidBody *doorBody = bpWorld.createRigidBody(1.0, // 質量 
                                                  transform,
                                                  doorShape);

  transform.setIdentity();
  transform.setOrigin(btVector3(0.5*D, 0.5*H + OFFSET, 0.5*D)); // 初期位置 (X, Y, Z)
  btRigidBody *jambBody = bpWorld.createRigidBody(0, // 質量(0 --> 固定)
                                                  transform,
                                                  jambShape);

  
  btVector3 doorPivot(0.5*W, 0, 0.5*D); // ヒンジの回転軸の位置（doorBodyのローカル座標）
  btVector3 doorAxis(0, 1, 0);          // ヒンジの回転軸の方向（doorBodyのローカル座標）
  btVector3 jambPivot(-0.5*D, 0, -0.5*D); // ヒンジの回転軸の位置（jambBodyのローカル座標）
  btVector3 jambAxis(0, 1, 0);            // ヒンジの回転軸の方向（jambBodyのローカル座標）
  g_joint = new btHingeConstraint(*doorBody, *jambBody,
                                  doorPivot, jambPivot,
                                  doorAxis,  jambAxis);
  g_joint->setLimit(-0.3*SIMD_PI, 0.3*SIMD_PI); // 下限，上限 [rad]（初期姿勢に対する相対値 ）
  g_joint->setDbgDrawSize(5); // 可動範囲表示の大きさ
  bpWorld.addConstraint(g_joint);
}

////////////////////////////////////////////////////////////////////////////////
/**
 *  プログラムの開始点
 */
int main(){
  // 動力学シミュレーションのワールドの準備
  BPWorld bpWorld("Door Hinge", // ウィンドウのタイトル
                  640, 480,         // ウィンドウの幅と高さ [pixels]
                  doMonitoring);    // モニタリング用関数へのポインタ
  
  setUpObjects(bpWorld); // 物体を作成して配置する
  
  bpWorld.enableConstraintDisplay(true); // ジョイントの可動範囲の表示を有効にする
  bpWorld.start(); // ウィンドウを表示して，シミュレーションを開始する
  
  return 0;
}
