////////////////////////////////////////////////////////////////////////////////
/**
 *  鉛直方向のマス・バネ・ダンパ系のシミュレーション
 *  （Hinge2Constraint の使用例）
 */

#include "BPWorld.h"
#include <iostream>

btRigidBody *g_movBoxBody; // 箱の力学データ

// #define EXEC_VALIDATION_EXP 

#ifdef EXEC_VALIDATION_EXP // 以下#ifdefブロック内は動作検証実験用の設定
#  include <fstream>
#  include <iomanip>
std::ofstream outf("result.txt"); // シミュレーション結果を記録するファイル
#endif


////////////////////////////////////////////////////////////////////////////////
/**
 *  シミュレーションのステップごとに呼ばれる関数（コールバック）
 *  物体の状態をモニタリングするために使う．
 */
void doMonitoring(btScalar currentTime){
#ifdef EXEC_VALIDATION_EXP // 以下#ifdefブロック内は動作検証実験用の設定    
  const btVector3& pos = g_movBoxBody->getCenterOfMassPosition();
  outf << std::setprecision(8) << std::showpoint;
  outf << currentTime << " " << pos.y() << '\n';
#endif  
}


////////////////////////////////////////////////////////////////////////////////
/**
 *  メッセージ msg を表示し，ユーザが入力した値を読む
 */
btScalar queryValue(const char *msg){
  using namespace std;
  btScalar val;
  cout << msg << flush;
  cin >> val;
  
  return val;
}


////////////////////////////////////////////////////////////////////////////////
/**
 *  物体を作成して，ワールドに配置する
 */
void setUpObjects(BPWorld& bpWorld){
  using namespace std;
  const btScalar W = 1; // 箱（立方体）の一辺の長さ
  const btScalar L = 5; // 平衡状態での二つの箱の中心間の距離
  btCollisionShape* boxShape = new btBoxShape(btVector3(0.5*W, 0.5*W, 0.5*W));
  btTransform transform;
  btScalar y0; // 初期変位（平衡位置からの変位）
  btScalar stiffness; // 剛性パラメータ（バネ係数）．フックの法則 F = kx の kに一致
  btScalar damping;   // 減衰パラメータ. 値の範囲 [0, 1]．1→減衰なし．0→バネ無効．
                      // 速度に比例する粘性摩擦とは異なる．
  btScalar mass = 1; // 自由端の箱の質量


  y0 = queryValue("Initial displacement (from the equilibrium point) y0 = ");
  stiffness = queryValue("stiffness = ");
  damping = queryValue("damping = ");

  // 固定端の箱を作る
  transform.setIdentity();
  transform.setOrigin(btVector3(0, 0.5*W, 0));
  btRigidBody *fixedBoxBody = bpWorld.createRigidBody(0, // 質量 0 -> 固定
                                                      transform,
                                                      boxShape);
  
  // 自由端の箱を作る
  transform.setOrigin(btVector3(0, L + 0.5*W, 0));
  g_movBoxBody = bpWorld.createRigidBody(mass, // 質量
                                         transform,
                                         boxShape);

  btVector3 axis1(0, 1, 0); // 回転軸1（ワールド座標） --> ローカルz軸
  btVector3 axis2(0, 0, 1); // 回転軸2（ワールド座標） --> ローカルx軸 
  btVector3 anchor // 2軸の交点（ワールド座標）
    = g_movBoxBody->getCenterOfMassPosition(); 
  btHinge2Constraint *joint = new btHinge2Constraint(*fixedBoxBody, *g_movBoxBody,
                                                     anchor,  
                                                     axis1,
                                                     axis2);

  // コンストラクタ内で以下のように設定される（これ以外はbtGeneric6DofSpringConstraintと同じ）
  //  - 現在の位置がバネの平衡位置
  //    --> ただし，このプログラムでは重力が働くので，実際の平衡位置は
  //        mg/k だけ鉛直下方にずれる．（m:質量，g:重力加速度，k:バネ係数）
  //        cf.: 運動方程式  m d^2y/dt^2 = -ky -mg = -k(y + mg/k)
  //  - ローカルz軸方向の可動範囲が [-1, 1]．x, y軸方向については固定
  //  - ローカルz軸方向のみバネが有効．バネ係数のデフォルト値は(2π)^2 ≒ 39.478418．
  //    --> 共振周波数 f = √(k/m)/ (2π) = 1 Hz (k = (2π)^2, m = 1)
  //  - 回転可動範囲: ローカルx軸まわり:[-π,+π]，y軸まわりは固定, z軸まわり: [-π/4,π/4]
  const int DOF_ID_TRANS_Z = 2;  // ローカルZ軸方向の自由度を示す番号
  joint->setStiffness(DOF_ID_TRANS_Z, stiffness);  // バネ係数を設定する
  joint->setDamping(DOF_ID_TRANS_Z, damping); // 減衰パラメータを設定する （デフォルト値0.01）

  joint->setLinearLowerLimit(btVector3(0, 0, -L));
  joint->setLinearUpperLimit(btVector3(0, 0, +L));
  joint->setAngularLowerLimit(btVector3(0, 0, -SIMD_PI));
  joint->setAngularUpperLimit(btVector3(0, 0, +SIMD_PI));
  
  // ジョイントをワールドに追加する
  bpWorld.addConstraint(joint);
  
  joint->setDbgDrawSize(5); // 可動範囲を表す軸，円弧の表示サイズ
  
  // 箱の位置を平衡位置からずらす
  g_movBoxBody->translate(btVector3(0, y0, 0));


  // 以下二つの処理はシミュレーション実行中はBPWorldによって自動的に行われるが，
  // 最初にシミュレーションを一時停止状態にしてCG表示を開始するため，ここでは明示的に行う．
  
  // motion stateを更新する（g_movBoxBody->getWorldTransform()に一致させる）
  //     デモの仕組み上，motion state を参照してCG描画が行われるので，このようにしないと内部状態と
  //     CGが一致しない．
  //    （下記をコメントアウトすると，箱が変位y0でなく変位ゼロの状態で表示される）
  g_movBoxBody->getMotionState()->setWorldTransform(g_movBoxBody->getWorldTransform());

  // 物体の位置とジョイントの計算の基準となる座標系を更新する．
  // （下記をコメントアウトすると，上記座標系が変位y0でなく変位ゼロの状態で表示される）
  //   cf. btDiscreteDynamicsWorld::debugDrawWorld()
  joint->calculateTransforms();

  // 速度が小さい状態が続いてもsleeping状態にならないようにする
  g_movBoxBody->forceActivationState(DISABLE_DEACTIVATION);

  // カメラ位置・姿勢の設定
  bpWorld.setCamera(0, 2.74, 0.99, // 注視点の位置 [x, y, z]
                    8, // 注視点からカメラ視点までの距離
                    0, // 方位角（azimuth） [deg]
                    12, // 仰角（elevation） [deg]
                    0, 1, 0); // 天井方向ベクトル [x, y, z] （cf. OpenGLのgluLookAt()）
}


////////////////////////////////////////////////////////////////////////////////
/**
 *  プログラムの開始点
 */
int main(){
  // 動力学シミュレーションのワールドの準備
  BPWorld bpWorld("Mass Spring Damper (Vertical)",  // ウィンドウのタイトル
                  640, 480,     // ウィンドウの幅と高さ [pixels]
                  doMonitoring); // モニタリング用関数へのポインタ

#ifdef EXEC_VALIDATION_EXP // 以下#ifdefブロック内は動作検証実験用の設定
  bpWorld.enableFixedTimeStep(true); // 一定の時間刻みでシミュレーションを進める
  bpWorld.setTimeStep(0.01); // 時間刻み
#endif  
  
  bpWorld.setGlobalERP(0.2);
  setUpObjects(bpWorld); // 物体を作成して配置する
  
  bpWorld.enableConstraintDisplay(true); // ジョイントの可動範囲の表示を有効にする
  bpWorld.toggleIdle(); // 一時停止状態から始める
  std::cout << "Focus on the CG window and then press 'i' to start/stop the simulation" << std::endl;
  
  bpWorld.start(); // ウィンドウを表示して，シミュレーションを開始する

  return 0;
}
