////////////////////////////////////////////////////////////////////////////////
/**
 *  水平方向のマス・バネ・ダンパ系のシミュレーション
 *
 */

#include "BPWorld.h"
#include <iostream>

btRigidBody *g_movBoxBody; // バネの自由端の箱の力学データ

// #define EXEC_VALIDATION_EXP

#ifdef EXEC_VALIDATION_EXP // 以下#ifdefブロック内は動作検証実験用の設定
#  include <fstream>
#  include <iomanip>
std::ofstream outf("result.txt"); // シミュレーション結果を記録するファイル
#endif

////////////////////////////////////////////////////////////////////////////////
/**
 *  シミュレーションのステップごとに呼ばれる関数（コールバック）
 *  物体の状態をモニタリングするために使う．
 */
void doMonitoring(btScalar currentTime){
#ifdef EXEC_VALIDATION_EXP // 以下#ifdefブロック内は動作検証実験用の設定
  const btVector3& pos = g_movBoxBody->getCenterOfMassPosition();
  outf << std::setprecision(8) << std::showpoint;
  outf << currentTime << " " << pos.x() << '\n';
#endif
}


////////////////////////////////////////////////////////////////////////////////
/**
 *  メッセージ msg を表示し，ユーザが入力した値を読む
 */
btScalar queryValue(const char *msg){
  using namespace std;
  btScalar val;
  cout << msg << flush;
  cin >> val;

  return val;
}

////////////////////////////////////////////////////////////////////////////////
/**
 *  物体を作成して，ワールドに配置する
 */
void setUpObjects(BPWorld& bpWorld){
  using namespace std;
  const btScalar W = 1; // 箱（立方体）の一辺の長さ
  const btScalar L = 5; // 平衡状態での二つの箱の中心間の距離
  btCollisionShape* boxShape = new btBoxShape(btVector3(0.5*W, 0.5*W, 0.5*W));
  btTransform transform;
  btScalar x0; // バネの平衡位置からの初期変位
  btScalar stiffness; // 剛性パラメータ（バネ係数）．フックの法則 F = kx の kに一致
  btScalar damping;   // 減衰パラメータ．値の範囲 [0, 1]．1→減衰なし．0→バネ無効．
                      // 速度に比例する粘性摩擦とは異なる．
  btScalar mass = 1; // 自由端の箱の質量

  x0 = queryValue("initial displacement (from the equilibrium point) x0 = ");
  stiffness = queryValue("stiffness = ");
  damping = queryValue("damping = ");

  // 共振周波数（理論値）√(k/m) /(2π)
  cout << "theoretical resonant freq.  = "
       << btSqrt(stiffness/mass)/(2*SIMD_PI) << endl;

  // 固定端の箱を作る
  transform.setIdentity();
  transform.setOrigin(btVector3(-L, 0.5*W, 0));
  btRigidBody *fixedBoxBody = bpWorld.createRigidBody(0, // 質量 0 -> 固定
                                                      transform,
                                                      boxShape);

  // 自由端の箱を作る
  transform.setOrigin(btVector3(0, 0.5*W, 0));
  g_movBoxBody = bpWorld.createRigidBody(mass, // 質量
                                         transform,
                                         boxShape);
  g_movBoxBody->setFriction(0.0);

  // ジョイントの計算の基準となる座標系（物体座標系を基準とするローカル座標系）
  //  btGeneric6DofSpringConstraintでは6自由度分の基準を一度に指定するため，
  //  コンストラクタの引数にbtTransformで基準とする座標系を指定する。
  btTransform trFixBox, trMovBox;
  trFixBox.setIdentity(); // fixedBoxBodyの物体座標系と一致するローカル座標系
  trMovBox.setIdentity(); // g_movBoxBodyの物体座標系と一致するローカル座標系

  // 二つの箱をつなぐジョイントを作る
  btGeneric6DofSpringConstraint *joint =
    new btGeneric6DofSpringConstraint(*fixedBoxBody, *g_movBoxBody,
                                      trFixBox, trMovBox,
                                      true); // true -> 第1引数の物体の位置姿勢を内部計算上の基準にする
                                             // false -> 第2引数の物体について同上）

  // ジョイントをワールドに追加する
  bpWorld.addConstraint(joint);

  const int DOF_ID = 0; // バネ，減衰に関する設定を行う対象の自由度を示す番号
                        // 番号と自由度の対応は次の通り
                        // 0: x軸方向平行移動, 1: y軸方向平行移動, 2: z軸方向平行移動
                        // 3: x軸まわりの回転, 4: y軸まわりの回転, 5: z軸まわりの回転
  joint->enableSpring(DOF_ID, true); // バネを有効にする（デフォルトでは無効）
  joint->setStiffness(DOF_ID, stiffness);  // バネ係数を設定する
  joint->setDamping(DOF_ID, damping); // 減衰パラメータを設定する

  // 可動範囲の設定（ローカル座標）
  // jointにより接続された物体の可動範囲を制限する
  joint->setLinearLowerLimit(btVector3(-100, 0, 0));
  joint->setLinearUpperLimit(btVector3( 100, 0, 0));
  joint->setAngularLowerLimit(btVector3(0, 0, 0));
  joint->setAngularUpperLimit(btVector3(0, 0, 0));

  // 現在の位置が平衡位置（バネが自然長となる位置）となるよう設定する
  joint->setEquilibriumPoint();

  joint->setDbgDrawSize(5); // 可動範囲を表す軸，円弧の表示サイズ

  // 箱の位置を平衡位置からずらす．
  g_movBoxBody->translate(btVector3(x0, 0, 0));


  // 以下二つの処理はシミュレーション実行中はBPWorldによって自動的に行われるが，
  // 最初にシミュレーションを一時停止状態にしてCG表示を開始するため，ここでは明示的に行う．

  // motion stateを更新する（g_movBoxBody->getWorldTransform()に一致させる）
  //     デモの仕組み上，motion state を参照してCG描画が行われるので，このようにしないと内部状態と
  //     CGが一致しない．
  //    （下記をコメントアウトすると，箱が変位y0でなく変位ゼロの状態で表示される）
  g_movBoxBody->getMotionState()->setWorldTransform(g_movBoxBody->getWorldTransform());

  // 物体の位置とジョイントの計算の基準となる座標系を更新する．
  // （下記をコメントアウトすると，上記座標系が変位y0でなく変位ゼロの状態で表示される）
  //   cf. btDiscreteDynamicsWorld::debugDrawWorld()
  joint->calculateTransforms();

   // 速度が小さい状態が続いてもsleeping状態にならないようにする
  g_movBoxBody->forceActivationState(DISABLE_DEACTIVATION);

  // カメラ位置・姿勢の設定
  bpWorld.setCamera(0, 2.74, 0.99, // 注視点の位置 [x, y, z]
                    8, // 注視点からカメラ視点までの距離
                    0, // 方位角（azimuth） [deg]
                    35, // 仰角（elevation） [deg]
                    0, 1, 0); // 天井方向ベクトル [x, y, z] （cf. OpenGLのgluLookAt()）
}


////////////////////////////////////////////////////////////////////////////////
/**
 *  プログラムの開始点
 */
int main(){
  // 動力学シミュレーションのワールドの準備
  BPWorld bpWorld("Mass Spring Damper",  // ウィンドウのタイトル
                  640, 480,     // ウィンドウの幅と高さ [pixels]
                  doMonitoring); // モニタリング用関数へのポインタ

#ifdef EXEC_VALIDATION_EXP // 以下#ifdefブロック内は動作検証実験用の設定
  bpWorld.enableFixedTimeStep(true); // 一定の時間刻みでシミュレーションを進める
  bpWorld.setTimeStep(0.01); // 時間刻み
#endif

  bpWorld.setGlobalERP(0.8);
  setUpObjects(bpWorld); // 物体を作成して配置する

  bpWorld.enableConstraintDisplay(true); // ジョイントの可動範囲の表示を有効にする
  bpWorld.toggleIdle(); // 一時停止状態から始める
  std::cout << "Focus on the CG window and then press 'i' to start/stop the simulation" << std::endl;

  bpWorld.start(); // ウィンドウを表示して，シミュレーションを開始する
  return 0;
}
