////////////////////////////////////////////////////////////////////////////////
/**
 *  Hinge2Constraint の使用例
 *
 */

#include "BPWorld.h"
#include <iostream>

btRigidBody *g_boxABody; // 箱Aの力学データ
btRigidBody *g_boxBBody; // 箱Bの力学データ

btHinge2Constraint *g_joint; // 箱A，Bをつなぐジョイント
btRotationalLimitMotor *g_motor1; // ジョイントの軸 axis1のモータ
btRotationalLimitMotor *g_motor2; // ジョイントの軸 axis2のモータ

btScalar g_refAng1 = 0; // モータ1の目標角度 [deg]
btScalar g_refAng2 = 0; // モータ2の目標角度 [deg]

////////////////////////////////////////////////////////////////////////////////
/**
 *  シミュレーションのステップごとに呼ばれる関数（コールバック）
 */
void doMonitoring(btScalar currentTime){
  using namespace std;
  const btScalar K = 0.8; // 比例ゲイン

  // 指定された目標角度になるように制御する
  g_motor1->m_targetVelocity = K*(btRadians(g_refAng1) - g_joint->getAngle1());
  g_motor2->m_targetVelocity = K*(btRadians(g_refAng2) - g_joint->getAngle2());

  cout << "reference: a1 = " << g_refAng1
       << ", a2 = " << g_refAng2 << "   "
       << "sensor: a1 = " << btDegrees(g_joint->getAngle1())
       << ", a2 = " << btDegrees(g_joint->getAngle2()) << "\n";
}


////////////////////////////////////////////////////////////////////////////////
/**
 *  キー入力処理
 */
void checkKey(int key, int x, int y){
  const btScalar DELTA = 1; // [deg]

  switch(key){
  case 'j':
    g_refAng1 += DELTA;
    break;
    
  case 'k':
    g_refAng1 -= DELTA;
    break;
    
  case '3':
    g_refAng2 += DELTA;
    break;
    
  case '4':
    g_refAng2 -= DELTA;
    break;
  }
}


////////////////////////////////////////////////////////////////////////////////
/**
 *  物体を作成して，ワールドに配置する
 */
void setUpObjects(BPWorld& bpWorld){
  const btScalar W = 1; // 箱A，B（立方体）の一辺の長さ
  btCollisionShape* boxShape = new btBoxShape(btVector3(0.5*W, 0.5*W, 0.5*W));
  btTransform transform;
  
  transform.setIdentity();
  transform.setOrigin(btVector3(0.5*W, 0.5*W, -0.5*W));
  g_boxABody = bpWorld.createRigidBody(0, // 質量
                                       transform,
                                       boxShape);

  transform.setOrigin(btVector3(-0.5*W, 1.5*W, 0.5*W));
  g_boxBBody = bpWorld.createRigidBody(0.1, // 質量  
                                       transform,
                                       boxShape);
  
  btVector3 axis1(0, 1, 0); // 回転軸1（ワールド座標）
  btVector3 axis2(0, 0, -1); // 回転軸2（ワールド座標）  
  btVector3 anchor(0, W, 0); // 2軸の交点（ワールド座標）
  g_joint = new btHinge2Constraint(*g_boxABody, *g_boxBBody,
                                   anchor,  
                                   axis1,
                                   axis2);

  // 可動範囲の設定（ローカル座標）--> 並進運動ができないように設定
  g_joint->setLinearLowerLimit(btVector3(0, 0, 0)); 
  g_joint->setLinearUpperLimit(btVector3(0, 0, 0));
  g_joint->setAngularLowerLimit(btVector3(-SIMD_PI, 0, -SIMD_PI)); 
  g_joint->setAngularUpperLimit(btVector3( SIMD_PI, 0,  SIMD_PI));
  
  bpWorld.addConstraint(g_joint);

  
  const int AXIS1_ID = 2; //
  const int AXIS2_ID = 0; //
  const btScalar MAX_MOTOR_TORQUE = 100;
  g_motor1 = g_joint->getRotationalLimitMotor(AXIS1_ID);
  g_motor2 = g_joint->getRotationalLimitMotor(AXIS2_ID);

  // モータの最大のトルクを指定する（メンバ名が"Force"となっているのはBPの都合）
  g_motor1->m_maxMotorForce = MAX_MOTOR_TORQUE; 
  g_motor2->m_maxMotorForce = MAX_MOTOR_TORQUE;
  
  // モータを有効にする
  g_motor1->m_enableMotor = true;
  g_motor2->m_enableMotor = true;

  // 停止が続いてもsleeping状態にならないようにする
  g_boxABody->forceActivationState(DISABLE_DEACTIVATION);
  g_boxBBody->forceActivationState(DISABLE_DEACTIVATION);
    
  // 可動範囲のCG表示サイズの指定
  g_joint->setDbgDrawSize(5);
}

////////////////////////////////////////////////////////////////////////////////
/**
 */
void showUsage()
{
  using namespace std;
  cout
    << "Usage: \n"
    << "   j/k: increase/decrease angle1 reference\n"
    << "   3/4: increase/decrease angle2 reference\n"
    << endl;
}

////////////////////////////////////////////////////////////////////////////////
/**
 *  プログラムの開始点
 */
int main(){
  // 動力学シミュレーションのワールドの準備
  BPWorld bpWorld("Simple Hinge2Constraint",  // ウィンドウのタイトル
                  640, 480,     // ウィンドウの幅と高さ [pixels]
                  doMonitoring, // モニタリング用関数へのポインタ
                  checkKey);    // キーボード入力処理用関数へのポインタ

  bpWorld.setTimeStep(0.01);
  bpWorld.setGlobalERP(0.2);
  bpWorld.enableConstraintDisplay(true); // ジョイントの可動範囲の表示を有効にする
  // bpWorld.toggleIdle(); // プログラム開始時に一時停止状態とする

  showUsage();
  setUpObjects(bpWorld); // 物体を作成して配置する
  bpWorld.start(); // ウィンドウを表示して，シミュレーションを開始する

  return 0;
}
