////////////////////////////////////////////////////////////////////////////////
/**
 *  三輪車のシミュレーション
 *
 */

#include "BPWorld.h" // BPWorldを使うため
#include <iostream>

BPWorld *g_bpWorld;
btRigidBody *g_driveWheelBody; // 駆動輪の力学データ
btRigidBody *g_carBody; // 車体力学データ
btHinge2Constraint *g_driveJoint; // 駆動輪と車体をつなぐジョイント
btRotationalLimitMotor *g_drive; // 駆動用モータ
btRotationalLimitMotor *g_steer; // ステアリング用モータ
btScalar g_refAngVel = 0; // 駆動輪の回転速度の目標値 [deg/s]
btScalar g_refSteerAng = 0; // ステアリング角の目標値 [deg]
bool g_trackingRequested = false; // カメラによる車体追従を行うか否かを表すフラグ

////////////////////////////////////////////////////////////////////////////////
/**
 *  シミュレーションのステップごとに呼ばれる関数（コールバック）
 *  物体の状態をモニタリングするために使う．
 */
void doMonitoring(btScalar currentTime){
  // 駆動輪の回転速度の制御 --> モータの速度制御機能を利用
  g_drive->m_targetVelocity = btRadians(g_refAngVel);
  
  // ステアリング角の制御 --> 比例制御
  // （直接目標角度を指定したいが，その機能がないため下記のように制御する）
  const btScalar K = 0.2; // 比例ゲイン
  g_steer->m_targetVelocity = K*(btRadians(g_refSteerAng) - g_driveJoint->getAngle1());

  if(g_trackingRequested){
    // カメラ位置・姿勢の調整: 車体の中心に追従する 
    btVector3 pc = g_carBody->getCenterOfMassPosition(); // 車体の中心
  
    // 車体ローカル座標系のx軸正方向の方向ベクトルを取得
    btVector3 xdir = btMatrix3x3(g_carBody->getOrientation()).getColumn(0);

    // xdirとワールド座標系Z軸のなす角を求める
    btScalar azimuth = btDegrees(atan2(xdir.x(), xdir.z()));
    g_bpWorld->setCamera(pc.x(), pc.y(), pc.z(), // 注視点の位置 [x, y, z]
                         15, // 注視点からカメラ視点までの距離
                         azimuth, // 方位角（azimuth） [deg]
                         13.2, // 仰角（elevation） [deg]
                         0, 1, 0); // 天井方向ベクトル [x, y, z] （cf. OpenGLのgluLookAt()）
  }
}

////////////////////////////////////////////////////////////////////////////////
/**
 *  キー入力処理
 */
void checkKey(int key, int x, int y){
  using namespace std;
  const btScalar DELTA = 5;
  const btScalar DELTA_ST = 0.5;

  switch(key){
  case 'j':
    g_refAngVel -= DELTA;
    break;
    
  case 'k':
    g_refAngVel += DELTA;
    break;

  case '3':
    g_refSteerAng += DELTA_ST;
    break;

  case '4':
    g_refSteerAng -= DELTA_ST;
    break;
    
  case 'T':
    g_trackingRequested = !g_trackingRequested;
    break;
    
  case '0':
    g_refAngVel = 0;
    g_refSteerAng = 0;
    break;
  }
  
  cout << "steering = " << g_refSteerAng << " deg, drive = " << g_refAngVel << " deg/s\n";
}

////////////////////////////////////////////////////////////////////////////////
/**
 *  物体を作成して，ワールドに配置する 
 */
void setUpObjects(BPWorld& bpWorld){
  const btScalar R_W = 1; // 車輪の半径
  const btScalar W_W = 1.2; // 車輪の幅
  const btScalar L_C = 6; // 車体の長さ
  const btScalar W_C = 4; // 車体の幅
  const btScalar D_C = 0.5; // 車体の厚さ
  const btScalar L_W = 4; // 軸距（wheelbase）: 前後の車輪間距離
  const btScalar L_A = 2.5; // 輪距（axle track）: 左右の車輪間距離
  const btScalar L_CW = 0.5*D_C + R_W + 0.1; // 車体重心を通る水平線から車輪中心までの距離
  btVector3 driveJointAnchor(0.5*L_W, R_W, 0); // 駆動輪の回転軸を置く点（ワールド座標）
  btVector3 psvLAnchor(-0.5*L_W, R_W, -0.5*L_A); // 右受動輪の回転軸を置く点（ワールド座標）
  btVector3 psvRAnchor(-0.5*L_W, R_W, +0.5*L_A); // 左受動輪の回転軸を置く点（ワールド座標）
  
  // 車体の形状データ
  btCollisionShape* carShape   = new btBoxShape(btVector3(0.5*L_C, 0.5*D_C, 0.5*W_C));
  // 車輪の形状データ
  btCollisionShape* wheelShape = new btCylinderShapeZ(btVector3(R_W, R_W, 0.5*W_W)); 
  
  btTransform transform;
  
  transform.setIdentity(); // 上記transformが作成された時点では
                           // その値（origin, rotation）が不定のため初期化する
  transform.setOrigin(btVector3(0, R_W + L_CW, 0));
  g_carBody = bpWorld.createRigidBody(2.0, // 質量 
                                      transform,
                                      carShape);

  // 駆動輪の作成
  transform.setOrigin(driveJointAnchor); 
  btRigidBody *g_driveWheelBody = bpWorld.createRigidBody(0.1, // 質量
                                                          transform,
                                                          wheelShape);
  g_driveWheelBody->setFriction(1);
  g_driveWheelBody->setRestitution(0);

  
  // 右受動輪の作成
  transform.setOrigin(psvRAnchor);
  btRigidBody *psvRWheelBody = bpWorld.createRigidBody(0.1, // 質量
                                                       transform,
                                                       wheelShape);
  psvRWheelBody->setFriction(1);   
  psvRWheelBody->setRestitution(0);

  
  // 左受動輪の作成
  transform.setOrigin(psvLAnchor);
  btRigidBody *psvLWheelBody = bpWorld.createRigidBody(0.1, // 質量
                                                       transform,
                                                       wheelShape);
  psvLWheelBody->setFriction(1);   
  psvLWheelBody->setRestitution(0);

  
  ////////////////////
  // ジョイントの作成
  
  // 駆動輪と車体をつなぐジョイント（2自由度）
  btVector3 axis1(0, -1, 0); // 回転軸1（ワールド座標）-> ステアリングの回転軸
  btVector3 axis2(0, 0, -1); // 回転軸2（ワールド座標）-> 駆動輪の回転軸
  g_driveJoint = new btHinge2Constraint(*g_carBody,
                                        *g_driveWheelBody,
                                        driveJointAnchor, // 2軸の交点
                                        axis1, axis2);
  // 上下方向の並進の可動範囲を設定
  //   回転軸1の方向（サスペンションあり）のデフォルト可動範囲が広いので狭める
  g_driveJoint->setLinearLowerLimit(btVector3(0, 0, -0.01)); // ローカル座標 
  g_driveJoint->setLinearUpperLimit(btVector3(0, 0, +0.01)); // ローカル座標
  bpWorld.addConstraint(g_driveJoint);


  // 右受動輪と車体をつなぐジョイント（1自由度）
  transform = g_carBody->getWorldTransform();
  btTransform wldToLoc = transform.inverse();  
  btVector3 pivotC = wldToLoc(psvRAnchor); // 車体上の車輪の取り付け位置（ローカル座標）
  btVector3 pivotW = btVector3(0, 0, 0); // 車輪の中心（車輪ローカル座標）
  btVector3 axisC = btVector3(0, 0, -1); // 車輪の回転軸（車体ローカル座標）
  btVector3 axisW = btVector3(0, 0, -1); // 車輪の回転軸（車輪ローカル座標）
  btHingeConstraint *psvRJoint = new btHingeConstraint(*g_carBody, *psvRWheelBody,
                                                       pivotC, pivotW,
                                                       axisC, axisW);
  bpWorld.addConstraint(psvRJoint);
  
  // 左受動輪と車体をつなぐジョイント（1自由度）
  pivotC = wldToLoc(psvLAnchor); // ワールド座標から車体ローカル座標への変換
  btHingeConstraint *psvLJoint = new btHingeConstraint(*g_carBody, *psvLWheelBody,
                                                       pivotC, pivotW,
                                                       axisC, axisW);
  bpWorld.addConstraint(psvLJoint);

  //（なお，上記wldToLoc()のような呼び出しで座標変換が行えるのはC++の演算子オーバロード機能が
  //　使われているため。BPのAPI Reference参照（btTransformのページ，operator()の項），
  //　http://bulletphysics.com/Bullet/BulletFull/classbtTransform.html#a56520e1c23a1ed7252cb0a9db7a3d60e ）



  ////////////////////
  // モータ使用準備（g_driveJointの作成時に各軸にモータが自動で同時に作成される）

  // g_driveJointの各軸のモータを区別する番号（ローカルのx，y，z軸について順に0，1，2）
  // （BPの都合上btHinge2Constraintの引数axis1, axis2と異なる番号になっている）
  const int DRIVE_ROT_AXIS = 0; // 駆動輪の回転軸（ローカル座標系x軸）
  const int STEER_ROT_AXIS = 2; // ステアリングの回転軸（ローカル座標系z軸）
  const btScalar MAX_MOTOR_TORQUE = 40;
  g_drive = g_driveJoint->getRotationalLimitMotor(DRIVE_ROT_AXIS);
  g_steer = g_driveJoint->getRotationalLimitMotor(STEER_ROT_AXIS);

  // モータの最大のトルクを指定する（メンバ名が"Force"となっているのはBPの都合）
  g_drive->m_maxMotorForce = MAX_MOTOR_TORQUE; 
  g_steer->m_maxMotorForce = MAX_MOTOR_TORQUE;

  // モータを有効にする
  g_drive->m_enableMotor = true;
  g_steer->m_enableMotor = true;


  // 停止が続いてもsleeping状態にならないようにする
  g_carBody->forceActivationState(DISABLE_DEACTIVATION);
  g_driveWheelBody->forceActivationState(DISABLE_DEACTIVATION);
  psvRWheelBody->forceActivationState(DISABLE_DEACTIVATION);
  psvLWheelBody->forceActivationState(DISABLE_DEACTIVATION);
}


////////////////////////////////////////////////////////////////////////////////
void showUsage(){
  using namespace std;
  cout
    << "Usage: \n"
    << "   j: drive forward\n"
    << "   k: drive backward\n"
    << "   3: turn left\n"
    << "   4: turn right\n"
    << "   T: enable/disable tracking camera\n"
    << "   0: reset references\n"
    << endl;
}


////////////////////////////////////////////////////////////////////////////////
/**
 *  プログラムの開始点
 */
int main(){
  // 動力学シミュレーションのワールドの準備
  BPWorld bpWorld("Tricycle", // ウィンドウのタイトル
                  640, 480,     // ウィンドウの幅と高さ [pixels]
                  doMonitoring, // モニタリング用関数へのポインタ
                  checkKey);    // キーボード入力処理用関数へのポインタ
  g_bpWorld = &bpWorld;
  
  setUpObjects(bpWorld); // 物体を作成して配置する
  showUsage();
  bpWorld.setGlobalERP(0.2);
  
  bpWorld.start(); // ウィンドウを表示して，シミュレーションを開始する
  
  return 0;
}
