////////////////////////////////////////////////////////////////////////////////
/**
 *  ロボットアームのシミュレーション
 *
 */

#include "BPWorld.h"
#include <iostream>
#include <deque>
#include <cstdlib>
#include <algorithm>

BPWorld *g_bpWorld;
btRigidBody *g_link1Body; //* リンク1
btRigidBody *g_link2Body; //* リンク2
const btScalar R_LINK = 0.2; //* リンクの半径
const btScalar L_LINK = 2;   //* リンクの長さ
const btVector3 g_tipPosLocal(0, 0.5* L_LINK + R_LINK, 0); //* リンクの先端位置（ローカル座標）

bool IGNORE_COLLISION = true; //* ジョイントで繋がれた物体同士の衝突を無視するか否かを表すフラグ
btUniversalConstraint *g_joint1; //* 台座とリンク1を繋ぐジョイント(2自由度)
btHingeConstraint     *g_joint2; //* リンク1とリンク2 を繋ぐジョイント(1自由度)

btScalar g_refAng1 = 0;   //* 目標角度1: リンク1回転軸1のまわりの回転角度 [deg]
btScalar g_refAng2 = 0;   //* 目標角度2: リンク1回転軸2のまわりの回転角度 [deg]
btScalar g_refAng3 = 0;   //* 目標角度3: リンク2回転軸のまわりの回転角度 [deg]

btRotationalLimitMotor *g_motor1; //* モータ1
btRotationalLimitMotor *g_motor2; //* モータ2
const btScalar MAX_MOTOR_TORQUE = 200; //* モータのトルクの最大値

// ピッキング（アームの先端で物を持ち上げる操作）の準備．対象物体: 円盤
// このプログラムでは，見た目上アーム先端に取り付けた磁石で対象物体を
// 吸着するようなピッキングを行う．
const btScalar R_DISK = 0.4; //* 円盤の半径
const btScalar H_DISK = 0.2; //* 円盤の厚さ
btCylinderShape *g_diskShape; //* 円盤の形状データ
std::deque<btRigidBody *> g_disks; //* 円盤の力学データ
btPoint2PointConstraint* g_pickConstraint = NULL; //* ピッキングに用いる一時的ジョイント

// キー入力処理のためのフラグ
bool g_newDiskRequested = false;
bool g_pickingRequested = false;
bool g_releaseRequested = false;
bool g_trackCamRequested = false;


////////////////////////////////////////////////////////////////////////////////
/**
 *  アームの先端の現在位置（ワールド座標）を返す
 */
inline btVector3 tipPosition(){
  // ローカル座標からワールド座標への変換（btTransformの関数呼出し式「()」演算子）
  return g_link2Body->getCenterOfMassTransform()(g_tipPosLocal);
}


////////////////////////////////////////////////////////////////////////////////
/**
 *  bodyAがbodyBよりもアームの先端に近いか否か調べる．
 *  この関数はC++標準テンプレートライブラリのmin_element()を利用するために
 *  定義する．詳細はC++言語についての書籍を参照．
 */
bool compareDistance(btRigidBody *bodyA, btRigidBody *bodyB){
  btVector3 tipPos = tipPosition();
  btScalar distA = btDistance(tipPos, bodyA->getCenterOfMassPosition());
  btScalar distB = btDistance(tipPos, bodyB->getCenterOfMassPosition());

  return distA < distB;
}


////////////////////////////////////////////////////////////////////////////////
/**
 *  円盤を作成し，半径Rの円上に分布するようにランダムに配置する．
 */
void addDiskAtRandom(BPWorld& bpWorld){
  const btScalar R = 3.5;
  const btScalar H = 2;
  btScalar th = SIMD_2_PI*((float)rand())/RAND_MAX; // 範囲[0, 2π]の疑似ランダム値
  btTransform transform;
  transform.setIdentity();
  transform.setOrigin(btVector3(R*btCos(th), H,  R*btSin(th))); // 半径Rの円上に置く

  btRigidBody *diskBody = bpWorld.createRigidBody(0.01, // 質量
                                                  transform, // 初期位置，姿勢
                                                  g_diskShape); // 形状データ
  diskBody->setFriction(0.8);
  g_disks.push_back(diskBody);

}

////////////////////////////////////////////////////////////////////////////////
/**
 *  ロボットアームの先端に最も近い円盤を持ち上げる．
 */
void pickUpDisk(){
  using namespace std;
  const btScalar THRESHOLD = 1.1; // この距離より先端-円盤間距離が遠ければピッキングを実行しない．
  btRigidBody *nearestDisk = *min_element(g_disks.begin(), g_disks.end(), compareDistance);
  btVector3 tipPos = tipPosition();
  btVector3 diskPos = nearestDisk->getCenterOfMassPosition();
  btScalar dist = btDistance(tipPosition(), nearestDisk->getCenterOfMassPosition());

  if(nearestDisk != NULL){
    cout << "Nearest disk at ("
         << diskPos.x() << ", " << diskPos.y() << ", " << diskPos.z() << ")" << endl;

    cout << "tip at ("
         << tipPos.x() << ", " << tipPos.y() << ", " << tipPos.z() << ")" << endl;
    cout << "distance: " << dist << endl;
  }

  if(nearestDisk != NULL &&
     g_pickConstraint == NULL && // ピッキング中の物体がないことを確認
     dist <= THRESHOLD){
    g_pickConstraint =
      new btPoint2PointConstraint(*g_link2Body,
                                  *nearestDisk,
                                  g_tipPosLocal,// アーム上の接続する点（ローカル座標）
                                  btVector3(0, 0.5*H_DISK, 0)); // 円盤上の接続する点（ローカル座標）
    g_pickConstraint->m_setting.m_tau = 0.6;
    g_pickConstraint->m_setting.m_damping = 0.1;

    g_bpWorld->addConstraint(g_pickConstraint); // IGNORE_COLLISION);

    cout << "Picking up a disk..." << endl;
  }else{
    cout << "Picking faild. Get the tip close to the disk more." << endl;
  }
}

////////////////////////////////////////////////////////////////////////////////
/**
 *  保持している円盤を離す
 */
void releaseDisk(){
  if(g_pickConstraint == NULL){
    return;
  }

  g_bpWorld->removeConstraint(g_pickConstraint);
  btRigidBody& disk = g_pickConstraint->getRigidBodyB();

  // 離したときに円盤が暴れないようにする
  disk.setLinearVelocity(btVector3(0, 0, 0));
  disk.setAngularVelocity(btVector3(0, 0, 0));

  // アーム先端に当たらないようにする
  disk.translate(btVector3(0, -(R_LINK + H_DISK), 0));

  // 円形の面が真上を向いた状態にする
  //disk.getOrientation() = btQuaternion::getIdentity();
  disk.getWorldTransform().setRotation(btQuaternion::getIdentity());
  // disk.getMotionState()->setWorldTransform(disk.getWorldTransform());

  // 加わっている力を無効にする
  // disk.clearForces();

  delete g_pickConstraint;
  g_pickConstraint = NULL;

}

////////////////////////////////////////////////////////////////////////////////
/**
 *  シミュレーションのステップごとに呼ばれる関数（コールバック）
 *  物体の状態をモニタリングするために使う．
 *
 *  @param currentTime 現在時刻
 */
void doMonitoring(btScalar currentTime){
  const btScalar K = 0.7; // 比例ゲイン

  // 各自由度についての比例制御
  g_motor1->m_targetVelocity = K*(btRadians(g_refAng1) - g_joint1->getAngle1());
  g_motor2->m_targetVelocity = K*(btRadians(g_refAng2) - g_joint1->getAngle2());
  g_joint2->enableAngularMotor(true,
                               K*(btRadians(g_refAng3) - g_joint2->getHingeAngle()),
                               MAX_MOTOR_TORQUE);

  //  btHingeConstraintは速度目標値のみを指定するためのメンバ関数を持っていないので
  //  上記 enableAngularMotor()を使用する

  // 要求があれば新たに円盤を追加する
  if(g_newDiskRequested){
    addDiskAtRandom(*g_bpWorld);
    g_newDiskRequested = false;
  }

  // ロボットアームによる円盤のピッキング
  if(g_pickingRequested){
    pickUpDisk();
    g_pickingRequested = false;
  }

  // 持ち上げた円盤が暴れないようにする
  if(g_pickConstraint != NULL){
    btRigidBody& disk = g_pickConstraint->getRigidBodyB();
    disk.setAngularVelocity(btVector3(0, 0, 0)); // 強制停止
  }

  // 円盤を手放す
  if(g_releaseRequested){
    releaseDisk();
    g_releaseRequested = false;
  }

  // カメラの位置・姿勢の設定
  if(g_trackCamRequested){
    btVector3 tipPos = tipPosition();
    g_bpWorld->setCamera(tipPos.x(), tipPos.y(), tipPos.z(),  // 注視点の位置 [x, y, z]
                         3,  // 注視点からカメラ視点までの距離
                         0,  // 方位角（azimuth） [deg]
                         15, // 仰角（elevation） [deg]
                         0, 1, 0); // 天井方向ベクトル [x, y, z] （cf. OpenGLのgluLookAt()）
  }
}

////////////////////////////////////////////////////////////////////////////////
btScalar clipValue(btScalar val, btScalar lowerLimit, btScalar upperLimit){
  if(val < lowerLimit){
    return lowerLimit;
  }
  if(val > upperLimit){
    return upperLimit;
  }
  return val;
}


////////////////////////////////////////////////////////////////////////////////
/**
 *  キー入力を処理する
 */
void checkKey(int key, int x, int y)
{
  using namespace std;
  const btScalar DELTA = 1; // [deg]

  switch(key){
  case '8':
    g_refAng1 += DELTA;
    break;
  case '9':
    g_refAng1 -= DELTA;
    break;

  case 'j':
    g_refAng2 += DELTA;
    break;
  case 'k':
    g_refAng2 -= DELTA;
    break;

  case '3':
    g_refAng3 += DELTA;
    break;
  case '4':
    g_refAng3 -= DELTA;
    break;

  case '@':
    g_newDiskRequested = true;
    break;

  case ',':
    g_pickingRequested = true;
    break;

  case '0':
    g_releaseRequested = true;
    break;

  case 'T':
    g_trackCamRequested = !g_trackCamRequested;
    break;
  }

  g_refAng1 = clipValue(g_refAng1, -180, +180); // 角度下限, 上限 [deg]
  g_refAng2 = clipValue(g_refAng2,  -90,  +90); // 角度下限, 上限 [deg]
  g_refAng3 = clipValue(g_refAng3,  -90,  +90); // 角度下限, 上限 [deg]

  cout << "references: "
       << g_refAng1 <<  ", " << g_refAng2 <<  ", " << g_refAng3 <<  " deg" << endl;
}


////////////////////////////////////////////////////////////////////////////////
/**
 *  物体を作成して，ワールドに配置する
 */
void setUpObjects(BPWorld& bpWorld){
  const btScalar H = 1; // 台座の高さ
  btCapsuleShape *linkShape = new btCapsuleShape(R_LINK, L_LINK);
  btRigidBody *baseBody;

  btTransform transform; // 座標変換データ
  transform.setIdentity();

  // 台座の作成
  transform.setOrigin(btVector3(0, 0.5*H, 0));
  baseBody = bpWorld.createRigidBody(0, // 質量 ゼロ --> 固定
                                     transform,  // 初期位置，姿勢
                                     new btCylinderShape(btVector3(R_LINK, 0.5*H, R_LINK))); // 形状データ


  // リンク1の作成
  transform.setOrigin(btVector3(0, 0.5*L_LINK + R_LINK + H, 0));
  g_link1Body = bpWorld.createRigidBody(0.1,        // 質量
                                        transform,  // 初期位置，姿勢
                                        linkShape); // 形状データ
  g_link1Body->setFriction(0.7);

  // リンク2の作成
  transform.setOrigin(transform.getOrigin() + btVector3(0, L_LINK + 2*R_LINK, 0));
  g_link2Body = bpWorld.createRigidBody(0.1,        // 質量
                                        transform,  // 初期位置，姿勢
                                        linkShape); // 形状データ
  g_link2Body->setFriction(0.7);




  btVector3 axis1(0, 1, 0); // 回転軸1（ワールド座標） --> ローカルz軸
  btVector3 axis2(0, 0, 1); // 回転軸2（ワールド座標） --> ローカルy軸
  btVector3 anchor(0, H, 0); // 2軸の交点（ワールド座標）
  g_joint1 = new btUniversalConstraint(*baseBody, *g_link1Body,
                                       anchor,
                                       axis1, axis2);
  g_joint1->setAngularLowerLimit(btVector3(0, -SIMD_PI, -SIMD_PI));
  g_joint1->setAngularLowerLimit(btVector3(0, +SIMD_PI, +SIMD_PI));

  bpWorld.addConstraint(g_joint1, IGNORE_COLLISION);
  // 台座とリンクの衝突を避けようとすると
  //  案1 衝突しないような幾何形状にする or
  //  案2 ピボットを物体表面上から離れた位置に置く
  //  案3 つないだ物体どうしの衝突を無視する設定をする
  //      （IGNORE_COLLISION == true）
  // などが考えられるが，ここでは案3を選ぶ．

  btVector3 pivotLink1(0,   0.5*L_LINK + R_LINK, 0);
  btVector3 pivotLink2(0, -(0.5*L_LINK + R_LINK), 0);
  btVector3 axisLink1(0, 0, 1);
  btVector3 axisLink2(0, 0, 1);
  g_joint2 = new btHingeConstraint(*g_link1Body, *g_link2Body,
                                   pivotLink1, pivotLink2,
                                   axisLink1, axisLink2);
  bpWorld.addConstraint(g_joint2, IGNORE_COLLISION);


  const int AXIS1_ID = 2;
  const int AXIS2_ID = 1;
  // モータ使用準備
  g_motor1 = g_joint1->getRotationalLimitMotor(AXIS1_ID); // ローカルz軸 = axis1
  g_motor2 = g_joint1->getRotationalLimitMotor(AXIS2_ID); // ローカルy軸 = axis2

  g_motor1->m_maxMotorForce = MAX_MOTOR_TORQUE;
  g_motor2->m_maxMotorForce = MAX_MOTOR_TORQUE;

  g_motor1->m_enableMotor = true;
  g_motor2->m_enableMotor = true;


  // 停止が続いてもsleeping状態にならないようにする
  g_link1Body->forceActivationState(DISABLE_DEACTIVATION);
  g_link2Body->forceActivationState(DISABLE_DEACTIVATION);

  g_joint1->setDbgDrawSize(5); // 可動範囲を表す軸，円弧の表示サイズ
  g_joint2->setDbgDrawSize(5);


  // ピッキング対象の形状データ
  g_diskShape = new btCylinderShape(btVector3(R_DISK, 0.5*H_DISK, R_DISK));

  const int INIT_NUM_DISKS = 10;
  for(int i = 0; i < INIT_NUM_DISKS; i++){
    addDiskAtRandom(bpWorld);
  }

  // カメラの位置・姿勢の設定
  bpWorld.setCamera(0, 5, 2, // 注視点の位置 [x, y, z]
                    7,  // 注視点からカメラ中心までの距離
                    0,  // 方位角（azimuth） [deg]
                    15, // 仰角（elevation） [deg]
                    0, 1, 0); // 天井方向ベクトル [x, y, z] （cf. OpenGLのgluLookAt()）
}


////////////////////////////////////////////////////////////////////////////////
/**
 *  使用法を表示する
 */
void showUsage()
{
  using namespace std;
  cout << "8/9: rotate the 1st link around the 1st axis\n"
       << "j/k: rotate the 1st link around the 2nd axis\n"
       << "3/4: rotate the 2nd link\n"
       << "@:   add a disk in the world\n"
       << ",:   pick up the disk nearest to the arm tip\n"
       << "0:   release the disk held at the tip\n"
       << "T:   toggle tracking camera mode\n"
       << endl;
}


////////////////////////////////////////////////////////////////////////////////
/**
 *  プログラムの開始点
 */
int main(){
  // 動力学シミュレーションのワールドの準備
  BPWorld bpWorld("Robot Arm", // ウィンドウのタイトル
                  640, 480,        // ウィンドウの幅と高さ [pixels]
                  doMonitoring,    // モニタリング用関数へのポインタ
                  checkKey);       // キーボード入力処理用関数へのポインタ

  g_bpWorld = &bpWorld;
  bpWorld.setGlobalERP(0.2);

  setUpObjects(bpWorld); // 物体を作成して配置する

  showUsage();

  bpWorld.enableConstraintDisplay(true); // ジョイントの可動範囲の表示を有効にする
  //bpWorld.toggleIdle(); // ウィンドウ表示の後，シミュレーションを一時停止状態にするため

  bpWorld.start(); // ウィンドウを表示して，シミュレーションを開始する

  return 0;
}
