////////////////////////////////////////////////////////////////////////////////
/**
 *  コマのシミュレーション
 */

#include "BPWorld.h" // BPWorldを使うため
#include <iostream>

////////////////////////////////////////////////////////////////////////////////
/**
 *  シミュレーションのステップごとに呼ばれる関数（コールバック）
 *  物体の状態をモニタリングするために使う．
 */
void doMonitoring(btScalar currentTime){

}

////////////////////////////////////////////////////////////////////////////////
/**
 *  物体を作成して，ワールドに配置する 
 */
void setUpObjects(BPWorld& bpWorld){
  btVector3 aabbMin;
  btVector3 aabbMax;
  btQuaternion quat;
  btScalar margin = 0.0;
  btCollisionShape* topShape = new btConeShape(1, 2.2); // 円錐の半径，高さ

  btTransform trans1;
  btTransform trans2;
  btTransform trans3;
  trans1.setIdentity();

  // 円錐の中心から頂点までの距離を調べる
  // --> aabbMax.y()がその距離．円錐の中心軸がY軸に一致．
  // --> "Aabb"は"Axis-Aligned Bounding Box"の省略
  //      (X,Y,Z軸に平行な面を持つ外接直方体)
  topShape->getAabb(trans1, aabbMin, aabbMax);

  // 頂点を床方向に向ける（準備だけで物体への適用は後）
  quat.setEuler(0, btRadians(180), 0); // 回転角度[rad]はY,X,Z軸まわりの順序
  trans1.setRotation(quat); // ワールド座標系の原点を基準とする回転
  // 頂点が床に付くように平行移動（これがゼロだと円錐は床に突き刺さる）
  trans1.setOrigin(btVector3(0, aabbMax.y() + margin, 0));

  // コマの傾き（頂点まわりの回転）の設定
  trans2.setIdentity();
  quat.setEuler(0, 0, btRadians(15)); // 回転角度[rad]はY,X,Z軸まわりの順序
  trans2.setRotation(quat);

  // 初期位置・姿勢は一つの座標変換（btTransform）で設定するため，二つの座標変換を合成する
  // --> trans1のほうが先，trans2のほうが後に適用される
  // --> デフォルトのカメラ初期状態だとワールド座標系（右手系）の軸の向きは次の通り
  //      X軸: 左方向
  //      Y軸: 上方向
  //      Z軸: 画面手前から奥に向かう方向
  trans3.mult(trans2, trans1);

  // 円錐の形状を持つ剛体（=コマ）を作って配置する．ここではじめて座標変換が適用される．
  btRigidBody *topBody = bpWorld.createRigidBody(1, // 質量
                                                 trans3,
                                                 topShape);
  topBody->setRestitution(0.2);
  topBody->setFriction(0.9);

  // コマの回転速度の初期値を指定する
  // --> コマの傾けた中心軸まわりの回転を設定する（基準はワールド座標系）
  trans3.setOrigin(btVector3(0, 0, 0)); // 平行移動成分は不要なのでゼロにする

  // trans(vec)でベクトルvecが座標変換transによって変換される
  topBody->setAngularVelocity(trans3(btVector3(0, 5*SIMD_PI, 0)));


  // カメラの設定
  bpWorld.setCamera(0, 2.74, 0.99, // 注視点の位置 [x, y, z]
                    8, // 注視点からカメラ視点までの距離
                    0, // 方位角（azimuth） [deg]
                    15, // 仰角（elevation） [deg]
                    0, 1, 0); // 天井方向ベクトル [x, y, z] （cf. OpenGLのgluLookAt()）
}

////////////////////////////////////////////////////////////////////////////////
/**
 *  プログラムの開始点
 */
int main(){
  // 動力学シミュレーションのワールドの準備
  BPWorld bpWorld("Spinning Top", // ウィンドウのタイトル
                  640, 480,         // ウィンドウの幅と高さ [pixels]
                  doMonitoring);    // モニタリング用関数へのポインタ
  
  setUpObjects(bpWorld); // 物体を作成して配置する

  bpWorld.toggleIdle(); // 一時停止状態から始める
  std::cout << "Press 'i' to start/stop the simulation" << std::endl;
  
  bpWorld.start(); // ウィンドウを表示して，シミュレーションを開始する
  
  return 0;
}
