////////////////////////////////////////////////////////////////////////////////
/**
 *  ジャイロスコープのシミュレーション
 *  
 */

#include "BPWorld.h"
#include <iostream>

BPWorld *g_bpWorld; //* 物体を配置するBPWorld
btRigidBody *g_driveWheelBody; //* 駆動輪の力学データ
btRigidBody *g_carBody; //* 車体の力学データ
btRigidBody *g_flywheelBody; //* ジャイロのフライホイールの力学データ
btHinge2Constraint *g_driveJoint; //* 駆動輪と車体をつなぐジョイント
btRotationalLimitMotor *g_drive; //* 駆動用モータ
btRotationalLimitMotor *g_steer; //* ステアリング用モータ
btScalar g_refAngVel = 0; //* 駆動輪の回転速度の目標値 [deg/s]
btScalar g_refSteerAng = 0; //* ステアリング角の目標値 [deg]
bool g_trackingRequested = false; //* カメラによる車体追従を行うか否かを表すフラグ
bool g_flywheelRotRequested = false;  //* フライホイールの回転要求を表すフラグ
bool g_flywheelStopRequested = false; //* フライホイールの停止要求を表すフラグ
const btScalar FLYWHEEL_ANG_VEL = 100*M_PI; //* フライホイールの角速度ベクトルのノルム
                                            // この値が大きいほど方向を維持する性能が高い

////////////////////////////////////////////////////////////////////////////////
/**
 *  シミュレーションのステップごとに呼ばれる関数（コールバック）
 *  物体の状態をモニタリングするために使う．
 */
void doMonitoring(btScalar currentTime){
  // 駆動輪の回転速度の制御 --> モータの速度制御機能を利用
  g_drive->m_targetVelocity = btRadians(g_refAngVel);
  
  // ステアリング角の制御 --> 比例制御
  // （直接目標角度を指定したいが，その機能がないため下記のように制御する）
  const btScalar K = 0.2; // 比例ゲイン
  g_steer->m_targetVelocity = K*(btRadians(g_refSteerAng) - g_driveJoint->getAngle1());

  // フライホイールの回転と強制停止の処理
  //   ジャイロスコープが方向を維持する現象が物理現象であり，制御によるものでない
  //   ことを明示するため，敢えてモータを使わない
  if(g_flywheelRotRequested){
    // フライホイールの回転軸（ローカル座標系のy軸）まわりの回転速度を与える
    // 解説: 
    //  setAngularVelocity()ではワールド座標で角速度ベクトルを指定するため，
    //  ワールド座標で表現されたフライホイールの回転軸の方向が必要。
    //  このため getOrientation()で得た姿勢（クォータニオン表現）と，
    //  その姿勢への回転変換をベクトルに適用する関数quatRotate()を使う。
    //  quatRotate()は変換後のベクトルを返す。
    // 参照:
    //    http://bulletphysics.com/Bullet/BulletFull/classbtRigidBody.html#a2bb57646e4a3e2a67ecd0e55d23cdad5
    //    http://bulletphysics.com/Bullet/BulletFull/btQuaternion_8h.html#a41643f41d42b592d640c7b33e9f66dd6
    
    g_flywheelBody->setAngularVelocity
      (quatRotate(g_flywheelBody->getOrientation(), btVector3(0, FLYWHEEL_ANG_VEL, 0)));
    
    g_flywheelRotRequested = false;
  }
  if(g_flywheelStopRequested){
    g_flywheelBody->setAngularVelocity(btVector3(0, 0, 0));
    g_flywheelStopRequested = false; // 毎ステップ強制停止すると力学的に従った動きにならないため
  }
    
  if(g_trackingRequested){
    // カメラ位置・姿勢の調整: 車体の中心に追従する 
    btVector3 pc = g_carBody->getCenterOfMassPosition(); // 車体の中心
  
    g_bpWorld->setCamera(pc.x(), pc.y(), pc.z(), // 注視点の位置 [x, y, z]
                         12, // 注視点からカメラ視点までの距離
                         90, // 方位角（azimuth） [deg]
                         20, // 仰角（elevation） [deg]
                         0, 1, 0); // 天井方向ベクトル [x, y, z] （cf. OpenGLのgluLookAt()）
  }
}


////////////////////////////////////////////////////////////////////////////////
/**
 *  キー入力処理
 */
void checkKey(int key, int x, int y){
  using namespace std;
  const btScalar DELTA = 5;
  const btScalar DELTA_ST = 0.5;

  switch(key){
  case 'j':
    g_refAngVel -= DELTA;
    break;
    
  case 'k':
    g_refAngVel += DELTA;
    break;

  case '3':
    g_refSteerAng += DELTA_ST;
    break;

  case '4':
    g_refSteerAng -= DELTA_ST;
    break;
    
  case 'T':
    g_trackingRequested = !g_trackingRequested;
    break;
    
  case '0':
    g_refAngVel = 0;
    g_refSteerAng = 0;
    break;

  case 'F':
    g_flywheelRotRequested  = true;
    g_flywheelStopRequested = false;
    break;
    
  case 'S':
    g_flywheelRotRequested  = false;
    g_flywheelStopRequested = true;
    break;
  }
  
  cout << "steering = " << g_refSteerAng << " deg, drive = " << g_refAngVel << " deg/s\n";
}


////////////////////////////////////////////////////////////////////////////////
/**
 *  三輪車を作成する。
 *  ジャイロの取り付けに関する部分以外は例題プログラムTricycleと同じ。
 *
 *  戻り値:  ジャイロを取り付ける物体の力学データ(baseBody)
 *
 *  引数
 *    bpWorld:  物体を作成，配置するBPワールド
 *    basePivot: (出力)外部ジンバルの底部を取り付ける位置（baseBodyのローカル座標）
 *    baseAxis:  (出力)外部ジンバルの軸の取り付け方向（baseBodyのローカル座標）
 */
btRigidBody* makeTricycleBody(BPWorld& bpWorld, btVector3& basePivot, btVector3& baseAxis){
  const btScalar R_W = 1;   // 車輪の半径
  const btScalar W_W = 1.2; // 車輪の幅
  const btScalar L_C = 6;   // 車体の長さ
  const btScalar W_C = 4;   // 車体の幅
  const btScalar D_C = 0.5; // 車体の厚さ
  const btScalar L_W = 4;   // 軸距（wheelbase）: 前後の車輪間距離
  const btScalar L_A = 2.5; // 輪距（axle track）: 左右の車輪間距離
  const btScalar L_CW = 0.5*D_C + R_W + 0.1; // 車体重心を通る水平線から車輪中心までの距離
  btVector3 driveJointAnchor(0.5*L_W, R_W, 0); // 駆動輪の回転軸を置く点（ワールド座標）
  btVector3 psvLAnchor(-0.5*L_W, R_W, -0.5*L_A); // 右受動輪の回転軸を置く点（ワールド座標）
  btVector3 psvRAnchor(-0.5*L_W, R_W, +0.5*L_A); // 左受動輪の回転軸を置く点（ワールド座標）
  
  // 車体の形状データ
  btCollisionShape* carShape   = new btBoxShape(btVector3(0.5*L_C, 0.5*D_C, 0.5*W_C));
  // 車輪の形状データ
  btCollisionShape* wheelShape = new btCylinderShapeZ(btVector3(R_W, R_W, 0.5*W_W)); 
  
  btTransform transform;
  
  transform.setIdentity(); // 上記transformが作成された時点では
                           // その値（origin, rotation）が不定のため初期化する
  transform.setOrigin(btVector3(0, R_W + L_CW, 0));
  g_carBody = bpWorld.createRigidBody(2.0, // 質量 
                                      transform,
                                      carShape);

  // 駆動輪の作成
  transform.setOrigin(driveJointAnchor); 
  btRigidBody *g_driveWheelBody = bpWorld.createRigidBody(0.1, // 質量
                                                          transform,
                                                          wheelShape);
  g_driveWheelBody->setFriction(1);
  g_driveWheelBody->setRestitution(0);

  
  // 右受動輪の作成
  transform.setOrigin(psvRAnchor);
  btRigidBody *psvRWheelBody = bpWorld.createRigidBody(0.1, // 質量
                                                       transform,
                                                       wheelShape);
  psvRWheelBody->setFriction(1);   
  psvRWheelBody->setRestitution(0);

  
  // 左受動輪の作成
  transform.setOrigin(psvLAnchor);
  btRigidBody *psvLWheelBody = bpWorld.createRigidBody(0.1, // 質量
                                                       transform,
                                                       wheelShape);
  psvLWheelBody->setFriction(1);   
  psvLWheelBody->setRestitution(0);

  
  ////////////////////
  // ジョイントの作成
  
  // 駆動輪と車体をつなぐジョイント（2自由度）
  btVector3 axis1(0, -1, 0); // 回転軸1（ワールド座標）-> ステアリングの回転軸
  btVector3 axis2(0, 0, -1); // 回転軸2（ワールド座標）-> 駆動輪の回転軸
  g_driveJoint = new btHinge2Constraint(*g_carBody,
                                        *g_driveWheelBody,
                                        driveJointAnchor, // 2軸の交点
                                        axis1, axis2);
  // 上下方向の並進の可動範囲を設定
  //   回転軸1の方向（サスペンションあり）のデフォルト可動範囲が広いので狭める
  g_driveJoint->setLinearLowerLimit(btVector3(0, 0, -0.01)); // ローカル座標 
  g_driveJoint->setLinearUpperLimit(btVector3(0, 0, +0.01)); // ローカル座標
  bpWorld.addConstraint(g_driveJoint);


  // 右受動輪と車体をつなぐジョイント（1自由度）
  transform = g_carBody->getWorldTransform();
  btTransform wldToLoc = transform.inverse();  
  btVector3 pivotC = wldToLoc(psvRAnchor); // 車体上の車輪の取り付け位置（ローカル座標）
  btVector3 pivotW = btVector3(0, 0, 0); // 車輪の中心（車輪ローカル座標）
  btVector3 axisC = btVector3(0, 0, -1); // 車輪の回転軸（車体ローカル座標）
  btVector3 axisW = btVector3(0, 0, -1); // 車輪の回転軸（車輪ローカル座標）
  btHingeConstraint *psvRJoint = new btHingeConstraint(*g_carBody, *psvRWheelBody,
                                                       pivotC, pivotW,
                                                       axisC, axisW);
  bpWorld.addConstraint(psvRJoint);
  
  // 左受動輪と車体をつなぐジョイント（1自由度）
  pivotC = wldToLoc(psvLAnchor); // ワールド座標から車体ローカル座標への変換
  btHingeConstraint *psvLJoint = new btHingeConstraint(*g_carBody, *psvLWheelBody,
                                                       pivotC, pivotW,
                                                       axisC, axisW);
  bpWorld.addConstraint(psvLJoint);

  //（なお，上記wldToLoc()のような呼び出しで座標変換が行えるのはC++の演算子オーバロード機能が
  //　使われているため。BPのAPI Reference参照（btTransformのページ，operator()の項），
  //　http://bulletphysics.com/Bullet/BulletFull/classbtTransform.html#a56520e1c23a1ed7252cb0a9db7a3d60e ）


  ////////////////////
  // モータ使用準備（g_driveJointの作成時に各軸にモータが自動で同時に作成される）

  // g_driveJointの各軸のモータを区別する番号（ローカルのx，y，z軸について順に0，1，2）
  // （BPの都合上btHinge2Constraintの引数axis1, axis2と異なる番号になっている）
  const int DRIVE_ROT_AXIS = 0; // 駆動輪の回転軸（ローカル座標系x軸）
  const int STEER_ROT_AXIS = 2; // ステアリングの回転軸（ローカル座標系z軸）
  const btScalar MAX_MOTOR_TORQUE = 400;
  g_drive = g_driveJoint->getRotationalLimitMotor(DRIVE_ROT_AXIS);
  g_steer = g_driveJoint->getRotationalLimitMotor(STEER_ROT_AXIS);

  // モータの最大のトルクを指定する（メンバ名が"Force"となっているのはBPの都合）
  g_drive->m_maxMotorForce = MAX_MOTOR_TORQUE; 
  g_steer->m_maxMotorForce = MAX_MOTOR_TORQUE;

  // モータを有効にする
  g_drive->m_enableMotor = true;
  g_steer->m_enableMotor = true;


  // 停止が続いてもsleeping状態にならないようにする
  g_carBody->forceActivationState(DISABLE_DEACTIVATION);
  g_driveWheelBody->forceActivationState(DISABLE_DEACTIVATION);
  psvRWheelBody->forceActivationState(DISABLE_DEACTIVATION);
  psvLWheelBody->forceActivationState(DISABLE_DEACTIVATION);

  
  //////////
  // ジャイロの取り付け位置，方向の設定
  basePivot = btVector3(0, 0.5*D_C, 0);
  baseAxis =  btVector3(0, 1, 0);
  return g_carBody;
}


////////////////////////////////////////////////////////////////////////////////
/**
 *  ジンバルの形状データを作成する。
 *  2種類の板それぞれ2枚ずつ（計4枚）と円筒2本を組み合わせてジンバルの形状を作る。
 *
 *        y軸↑                     ジンバル断面は正方形とする。                           
 *           ｜                     ジンバルの中心をローカル座標系の原点に置く。           
 *     ┌─────┐               シャフトの長軸の向きをローカル座標系z軸の向きとする。
 *     ｜┌───┐｜─────
 *  ↓ ｜｜  ｜O ｜｜         ↑
 *  ＝=┤｜  ┼─｜├=＝→ z軸｜内接正方形の一辺の長さ l
 *  ↑ ｜｜      ｜｜         ↓
 * 2*ra｜└───┘｜─────
 *     └─────┘
 *
 *    Fig. ジンバルの形状
 * 
 *
 *  引数
 *    l:  ジンバルの内接正方形の一辺の長さ
 *    w:  ジンバルの板の幅
 *    d:  ジンバルの板の厚さ
 *    la: ジンバルのシャフトの長さ
 *    ra: ジンバルのシャフトの半径
 *
 *  戻り値: 作成した形状データ
 *  
 */
btCollisionShape* makeGimbalShape(btScalar l, btScalar w, btScalar d, btScalar la, btScalar ra){
  const btScalar ll = l + 2*d; // 板1の長さ
  btCollisionShape *plate1Shape = new btBoxShape(btVector3(0.5*w, 0.5*ll, 0.5*d)); // 板1の形状データ
  btCollisionShape *plate2Shape = new btBoxShape(btVector3(0.5*w, 0.5*d,  0.5*l)); // 板2の形状データ
  btCollisionShape *axleShape  = new btCylinderShapeZ(btVector3(ra, ra, 0.5*la)); // シャフトの形状データ
  btCompoundShape *gimbalShape = new btCompoundShape(); // 内部ジンバルの形状データ（合成形状）

  btTransform trans; // 相対的な座標変換（gimbalShapeが定めるローカル座標系を基準とする）
  trans.setIdentity();

  // 4枚の板を組み合わせてジンバルの形状を作る
  trans.setOrigin(btVector3(0, 0, 0.5*(ll - d)));
  gimbalShape->addChildShape(trans, plate1Shape);
  
  trans.setOrigin(btVector3(0, 0, -0.5*(ll - d)));
  gimbalShape->addChildShape(trans, plate1Shape);
  
  trans.setOrigin(btVector3(0, 0.5*(ll - d), 0));
  gimbalShape->addChildShape(trans, plate2Shape);
  
  trans.setOrigin(btVector3(0,-0.5*(ll - d), 0));
  gimbalShape->addChildShape(trans, plate2Shape);

  // 2本の円筒でジンバルの回転軸部分の形状を作る
  trans.setOrigin(btVector3(0, 0, 0.5*(ll + la)));
  gimbalShape->addChildShape(trans, axleShape);

  trans.setOrigin(btVector3(0, 0, -0.5*(ll + la)));
  gimbalShape->addChildShape(trans, axleShape);

  return gimbalShape;
}


////////////////////////////////////////////////////////////////////////////////
/**
 *  ジャイロスコープを作成する。
 *
 *  ジャイロスコープは内側から次の部品で構成される。
 *   - フライホイール
 *   - 内部ジンバル
 *   - 外部ジンバル
 *  各部品は1自由度のジョイントにより接続される。
 *
 *  簡略化のため取り付け対象物体の回転（ジャイロを鉛直方向に対し斜めに取り付けるケース）
 *  は考慮しない。すなわち，常にワールド鉛直上方向に外部ジンバルが立った状態の
 *  ジャイロスコープを作成する。
 *
 *  引数
 *    bpWorld:  物体を作成，配置するBPワールド
 *    baseBody:  取り付け対象物体の力学データ
 *    basePivot: 外部ジンバルの底部を取り付ける位置（baseBodyのローカル座標）
 *    baseAxis:  外部ジンバルの軸の取り付け方向（baseBodyのローカル座標）
 */
void makeGyroscopeBody(BPWorld& bpWorld,
                       btRigidBody* baseBody, btVector3& basePivot, btVector3& baseAxis){
  const btScalar L_A = 2;       // フライホイール軸の長さ
  const btScalar R_A = 0.2;     // フライホイール軸の半径
  const btScalar L_W = 0.2*L_A; // フライホイール中央の円盤の厚さ
  const btScalar R_W = 4*R_A;   // フライホイール中央の円盤の半径
  const btScalar L_G1  = L_A; // 内部ジンバルの内接正方形の一辺の長さ
  const btScalar D_G1  = 0.1; // 内部/外部ジンバルの板の厚さ              
  const btScalar W_G1  = 0.4; // 内部/外部ジンバルの板の幅                
  const btScalar LA_G1 = 0.6; // 内部/外部ジンバルのシャフトの長さ        
  const btScalar RA_G1 = 0.2; // 内部/外部ジンバルのシャフトの半径
  const btScalar L_G2 = L_G1 + 2*D_G1 + 2*LA_G1; // 外部ジンバルの内接正方形の一辺の長さ
  btQuaternion quat;
  btTransform trans;
  trans.setIdentity();
  
  // フライホイールの形状データの作成（円盤と軸を組み合わせた形状）
  btCompoundShape *flywheelShape = new btCompoundShape();  
  flywheelShape->addChildShape(trans, // flywheelShapeのローカル座標系に対する相対的な座標変換
                               new btCylinderShape(btVector3(R_A, 0.5*L_A, 0))); // 軸の形状
  flywheelShape->addChildShape(trans, // flywheelShapeのローカル座標系に対する相対的な座標変換
                               new btCylinderShape(btVector3(R_W, 0.5*L_W, 0))); // 円盤の形状
  
  // 内部ジンバルの形状データ
  btCollisionShape *innerGimbalShape =
    makeGimbalShape(L_G1,   // ジンバルの内接正方形の一辺の長さ
                    W_G1,   // ジンバルの板の厚さ                    
                    D_G1,   // ジンバルの板の幅
                    LA_G1,  // ジンバルのシャフトの長さ
                    RA_G1); // ジンバルのシャフトの半径        
  
  // 外部ジンバルの形状データ
  btCollisionShape *outerGimbalShape =
    makeGimbalShape(L_G2,   // ジンバルの内接正方形の一辺の長さ
                    W_G1,   // ジンバルの板の厚さ                    
                    D_G1,   // ジンバルの板の幅
                    LA_G1,  // ジンバルのシャフトの長さ
                    RA_G1); // ジンバルのシャフトの半径


  ////////////////////////////////////////
  // 各部品を配置して，接続する
  
  // 車体上の取り付け位置basePivotに合わせてフライホイール，内部・外部ジンバルを配置する。
  // basePivot は車体のローカル座標なので，ワールド座標に変換してから位置計算をする。

  // 外部ジンバルの外接正方形の一辺の長さの半分（中心から回転軸の端までの距離）
  const btScalar OFFSET = 0.5*(L_G2 + 2*D_G1 + 2*LA_G1);
  
  btTransform locToWld = baseBody->getWorldTransform(); // ローカル→ワールド座標変換
  trans.setOrigin(locToWld(basePivot) + btVector3(0, OFFSET, 0));
  
  quat.setEuler(0, 0, btRadians(90)); // 回転角度[rad] （Y,X,Z軸の順序）
  trans.setRotation(quat);

  // フライホイールの力学データ
  g_flywheelBody = bpWorld.createRigidBody(0.5, // 質量
                                           trans,
                                           flywheelShape);
  

  // 内部ジンバルの力学データ
  btRigidBody *innerGimbalBody = bpWorld.createRigidBody(0.02, // 質量
                                                         trans,
                                                         innerGimbalShape);
  

  quat.setEuler(0, btRadians(90), 0); // 回転角度[rad] （Y,X,Z軸の順序）
  trans.setRotation(quat);
  
  // 外部ジンバルの力学データ
  btRigidBody *outerGimbalBody = bpWorld.createRigidBody(0.02, // 質量
                                                         trans,
                                                         outerGimbalShape);
  

  // Fw: flywheel, InG: inner gimbal
  btVector3 pivot1Fw(0, 0, 0);  // ピボット（フライホイールのローカル座標）
  btVector3 axis1Fw(0, 1, 0);   // 軸の方向（フライホイールのローカル座標）
  btVector3 pivot1InG(0, 0, 0); // ピボット（内部ジンバルのローカル座標）
  btVector3 axis1InG(0, 1, 0);  // 軸の方向（内部ジンバルのローカル座標）  
  
  // フライホイールと内部ジンバルをつなぐジョイント
  btHingeConstraint *hingeFwInG = new btHingeConstraint(*g_flywheelBody,
                                                        *innerGimbalBody,
                                                        pivot1Fw, pivot1InG,
                                                        axis1Fw,  axis1InG);
  bpWorld.addConstraint(hingeFwInG);
  hingeFwInG->setLimit(-SIMD_PI, +SIMD_PI);
  

  // OutG: outer gimbal
  btVector3 pivot2InG(0, 0, 0);  // ピボット（内部ジンバルのローカル座標）
  btVector3 axis2InG(0, 0, 1);   // 軸の方向（内部ジンバルのローカル座標）
  btVector3 pivot2OutG(0, 0, 0); // ピボット（外部ジンバルのローカル座標）  
  btVector3 axis2OutG(0, 1, 0);  // 軸の方向（外部ジンバルのローカル座標）  
  
  // 内部ジンバルと外部ジンバルをつなぐジョイント
  btHingeConstraint *hingeInOutG = new btHingeConstraint(*innerGimbalBody,
                                                         *outerGimbalBody,
                                                         pivot2InG, pivot2OutG,
                                                         axis2InG,  axis2OutG);
  bpWorld.addConstraint(hingeInOutG);
  hingeInOutG->setLimit(-SIMD_PI, +SIMD_PI);


  // 外部ジンバルはX軸周りに90度回転して配置していることに留意
  btVector3 pivot3OutG(0, 0, OFFSET); // ピボット（外部ジンバルのローカル座標）
  btVector3 axis3OutG(0, 0, -1);      // 軸の方向（外部ジンバルのローカル座標）  
  
  // 外部ジンバルと取り付け対象物体をつなぐジョイント
  btHingeConstraint *hingeOutGBase = new btHingeConstraint(*outerGimbalBody,
                                                           *baseBody,
                                                           pivot3OutG, basePivot,
                                                           axis3OutG,  baseAxis);
  bpWorld.addConstraint(hingeOutGBase);
  hingeOutGBase->setLimit(-SIMD_PI, +SIMD_PI);
  hingeOutGBase->setDbgDrawSize(5); // 可動範囲表示の大きさ
  

  // 摩擦の設定
  g_flywheelBody->setFriction(0.0);
  innerGimbalBody->setFriction(0.0);
  outerGimbalBody->setFriction(0.1);
  baseBody->setFriction(0.4);

  // 解説: 外部ジンバルと取付け対象物体の間に摩擦を設定する理由
  //
  //   フライホイール，ジンバルは軸に関して対称なので，両側の対称な部分に
  //   働く重力によるトルクは打ち消し合い，実効ゼロである。
  //   したがって，
  //   - 摩擦がないと → 回転に関してジャイロは他物体の影響を受けず，
  //                       慣性にしたがって同じ方向を向き続ける
  //   - 摩擦があると → 外部ジンバルに摩擦によるトルクτが働く。
  //   --> フライホイールの回転速度（角運動量L）がゼロだと，外部ジンバルが回転する
  //   --> フライホイールの回転速度（角運動量L）が十分に大きいと，
  //     摩擦はフライホイール回転軸を水平方向から傾けるように作用する
  //     （dL(t)/dt = τ → L(t+δt) = L(t) + τδt。
  //       ベクトルLとτの向きに留意   L: X軸正方向(初期値)， τ: 鉛直上 or 下向き）。
  //     一方，フライホイール回転軸は鉛直軸まわりには動かないので，
  //     指し示す方位（回転軸方向ベクトルの水平成分）は変化しない（ジャイロコンパスの原理）。

  
  //////////
  // 初期状態での運動の設定
  g_flywheelBody->setAngularVelocity(btVector3(FLYWHEEL_ANG_VEL, 0, 0));
  // g_flywheelBody->setAngularVelocity(btVector3(0, 0, 0));  // ゼロの場合

  g_flywheelBody->forceActivationState(DISABLE_DEACTIVATION);
  innerGimbalBody->forceActivationState(DISABLE_DEACTIVATION);
  outerGimbalBody->forceActivationState(DISABLE_DEACTIVATION);
}


////////////////////////////////////////////////////////////////////////////////
/**
 *  物体を作成して，ワールドに配置する 
 */
void setUpObjects(BPWorld& bpWorld){
  btVector3 basePivot; // 三輪車上のジャイロスコープ取り付け位置座標（ローカル座標）
  btVector3 baseAxis;  // 三輪車上のジャイロスコープ取り付け方向軸（ローカル座標）
  btRigidBody *baseBody; // ジャイロスコープを取り付る対象の物体の力学データ

  // 三輪車を作る
  baseBody = makeTricycleBody(bpWorld,
                              basePivot, baseAxis); // basePivot, baseAxisには計算結果が入る
  
  // ジャイロスコープを作る
  makeGyroscopeBody(bpWorld, baseBody, basePivot, baseAxis);
}


////////////////////////////////////////////////////////////////////////////////
/**
 *  プログラムの使用法を表示する
 */
void showUsage(){
  using namespace std;
  cout
    << "Usage: \n"
    << "   j: drive forward\n"
    << "   k: drive backward\n"
    << "   3: turn left\n"
    << "   4: turn right\n"
    << "   T: enable/disable tracking camera\n"
    << "   0: reset references\n"
    << "   F: spin the flywheel\n"
    << "   S: stop the flywheel\n"
    << endl;
}


////////////////////////////////////////////////////////////////////////////////
/**
 *  プログラムの開始点
 */
int main(){
  // 動力学シミュレーションのワールドの準備
  BPWorld bpWorld("Gyroscope",   // ウィンドウのタイトル
                  640, 480,      // ウィンドウの幅と高さ [pixels]
                  doMonitoring, // モニタリング用関数へのポインタ
                  checkKey);    // キーボード入力処理用関数へのポインタ
  g_bpWorld = &bpWorld;
  
  setUpObjects(bpWorld); // 物体を作成して配置する
  showUsage();
  
  bpWorld.setGlobalERP(0.2);
  // bpWorld.toggleIdle(); // 開始時にシミュレーションを一時停止状態にする

  bpWorld.enableConstraintDisplay(true); // ジョイントの可動範囲の表示を有効にする
  bpWorld.start(); // ウィンドウを表示して，シミュレーションを開始する
  
  return 0;
}
