////////////////////////////////////////////////////////////////////////////////
/**
 *  ロープのシミュレーション
 */

#include "BPWorld.h" 
#include <iostream>

////////////////////////////////////////////////////////////////////////////////
/**
 *  物体を作成して，ワールドに配置する
 */
void setUpObjects(BPWorld& bpWorld){
  const btScalar L = 4.5; // ロープの全長
  const btScalar H = 5;   // ロープの各節点の高さ(Y座標)の初期値
  btVector3 ps = btVector3(0, H, 0); // ロープの始点の位置（ワールド座標）
  btVector3 pe = btVector3(L, H, 0); // ロープの終点の位置（ワールド座標）
  const int NUM_NODES = 20; // ロープの節点の個数
  const int FIXEDS = 0x01; // ロープの両端の節点をワールド座標系に対して固定するか
                           // 否かを示すフラグ
                           //  第1ビット -> 始点 (1なら固定，0なら自由)
                           //  第2ビット -> 終点 (1なら固定，0なら自由)

  // ロープの作成
  //   柔軟体を表現するクラスはbtSoftBodyである．これのみで様々な形状を扱える．
  //   しかしながらその形状を指定するには個別に形状を記述しなければならず，煩雑である．
  //   そこでBulletPhysicsライブラリには典型的な形状に関して柔軟体を容易に
  //   作成できるようにbtSoftBodyHelpersというクラスを用意している．
  //   btSoftBodyHelpersのメンバ関数を用いること以下のような形状の柔軟物体を
  //   作成できる．
  //    - ロープ → btSoftBodyHelpers::CreateRope()
  //    - 矩形の布 → btSoftBodyHelpers::CreatePatch()
  //    - 楕円体 → btSoftBodyHelpers::CreateEllipsoid()
  //    - 三角メッシュデータで表現された形状 → btSoftBodyHelpers::CreateFromTriMesh()
  //    - 所与の点群に対する凸包 → btSoftBodyHelpers::CreateFromConvexHull()
  //       凸包とは所与の点全てを包含し，かつ凹み部分がないようにして作った最小の多面体/多角形
  //       例: ☆型に対する凸包は五角形
  btSoftBody *ropeBody = 
    btSoftBodyHelpers::CreateRope(bpWorld.getSoftBodyWorldInfo(), // 柔軟体の作成に必要な情報
                                  ps, // 始点の位置（ワールド座標）
                                  pe, // 終点の位置（ワールド座標）
                                  NUM_NODES - 2, // 節点の個数 - 2 （「-2」は両端の節点の分）
                                  FIXEDS);
  
  ropeBody->setTotalMass(1); // 全体の質量の指定
  bpWorld.addSoftBody(ropeBody);

  // カメラの位置・姿勢の指定
  bpWorld.setCamera(0, 3, 0, // 注視点の位置 [x, y, z]
                    8, // 注視点からカメラ視点までの距離
                    0, // 方位角（azimuth） [deg]
                    15, // 仰角（elevation） [deg]
                    0, 1, 0); // 天井方向ベクトル [x, y, z] （cf. OpenGLのgluLookAt()）
}


////////////////////////////////////////////////////////////////////////////////
/**
 *  使用法を表示する
 */
void showUsage()
{
  using namespace std;
  cout << " Press 'i' to start/pause the simulation" << endl;
}


////////////////////////////////////////////////////////////////////////////////
/**
 *  プログラムの開始点
 */
int main(){
  // 動力学シミュレーションのワールドの準備
  BPWorld bpWorld("Rope",   // ウィンドウのタイトル
                  640, 480,     // ウィンドウの幅と高さ [pixels]
                  BPWorld::SOFT_BODY_INCLUDED);
  
  setUpObjects(bpWorld); // 物体を作成して配置する
  showUsage();
  
  bpWorld.toggleIdle(); // ウィンドウ表示の後，シミュレーションを一時停止状態にするため
  bpWorld.start(); // ウィンドウを表示して，シミュレーションを開始する

  return 0;
}
