////////////////////////////////////////////////////////////////////////////////
/**
 *  ウィンチのシミュレーション
 */

#include "BPWorld.h" 
#include <iostream> 

btHingeConstraint *g_rotorJoint; //* 回転体の回転軸として機能するヒンジ
btScalar g_angVel = 0; //* 角速度 [deg/s]

////////////////////////////////////////////////////////////////////////////////
/**
 *  シミュレーションのステップごとに呼ばれる関数（コールバック）
 *  物体の状態をモニタリングするために使う．
 */
void doMonitoring(btScalar currentTime){
  const btScalar MAX_MOTOR_TORQUE = 5;
  g_rotorJoint->enableAngularMotor(true, btRadians(g_angVel), MAX_MOTOR_TORQUE);
}


////////////////////////////////////////////////////////////////////////////////
/**
 * キー入力処理
 */
void checkKey(int key, int x, int y){
  using namespace std;
  const btScalar DELTA = 5; // [deg/s]
  const btScalar LL_ANG_VEL = -180; // [deg/s] 下限
  const btScalar UL_ANG_VEL =  180; // [deg/s] 上限
  
  switch(key){
  case 'j':
    g_angVel += DELTA;
    break;
    
  case 'k':
    g_angVel -= DELTA;
    break;

  case '0':
    g_angVel = 0;
    break;
  }

  if(g_angVel > UL_ANG_VEL){
    g_angVel = UL_ANG_VEL;
  }
  if(g_angVel < LL_ANG_VEL){
    g_angVel = LL_ANG_VEL;
  }

  cout << "angVel = " << g_angVel << " deg/s\n";
}


////////////////////////////////////////////////////////////////////////////////
/**
 *  物体を作成して，ワールドに配置する
 */
void setUpObjects(BPWorld& bpWorld){
  const btScalar R = 0.5; // 回転体（円柱）の半径
  const btScalar D = 3;   // 回転体（円柱）の厚さ
  const btScalar B = 2;   // 持ち上げる箱の一辺の長さ
  const btScalar H = 20;  // 回転軸を固定する高さ
  const int NUM_NODES = 60; // ロープを構成する節点の個数
  const btScalar L = 15; // ロープの長さ
  
  // 回転体の形状データ
  btCylinderShape *rotorShape = new btCylinderShapeZ(btVector3(R, R, 0.5*D));

  // 箱の形状データ
  btBoxShape *boxShape = new btBoxShape(btVector3(0.5*B, 0.5*B, 0.5*B));

  btTransform transform; // 座標変換データ     
  transform.setIdentity(); // 座標変換データの初期化（回転なし，平行移動なし）
  
  // 回転体（ロープ巻き取り機）を作成する
  transform.setOrigin(btVector3(-R, H, 0));
  btRigidBody *rotorBody = bpWorld.createRigidBody(1.0,        // 質量
                                                   transform,  // 初期位置，姿勢
                                                   rotorShape); // 形状データ

  // 持ち上げる箱を作成する
  transform.setOrigin(btVector3(0, H - L, 0));
  btRigidBody *boxBody = bpWorld.createRigidBody(0.2, // 質量
                                                 transform, // 初期位置，姿勢
                                                 boxShape); // 形状データ

  btVector3 pivot(0, 0, 0); // ピボットの位置（ローカル座標
  btVector3 axis(0, 0, 1); // 回転軸の方向（ローカル座標）

  // ワールドに対してピボットを固定したヒンジを作る（回転は可能）
  g_rotorJoint =
    new btHingeConstraint(*rotorBody,
                          pivot,
                          axis);
  bpWorld.addConstraint(g_rotorJoint);

  // ロープ
  btSoftBody *ropeBody = btSoftBodyHelpers::CreateRope(bpWorld.getSoftBodyWorldInfo(),
                                                       btVector3(0, H, 0), // 始点の位置（ワールド座標）
                                                       btVector3(0, H - L, 0), // 終点の位置（ワールド座標）
                                                       NUM_NODES - 2,
                                                       0x00); // ワールドに対する固定を無効にする
  bpWorld.addSoftBody(ropeBody);
  
  ropeBody->setTotalMass(1);
  ropeBody->appendAnchor(0, rotorBody); // 始点を回転体に固定
  ropeBody->appendAnchor(NUM_NODES-1, boxBody); // 終点を箱に固定

  // 各節点間の剛性（linear stiffness coefficient）の設定．
  // 値の範囲 [0, 1]．1に近いほど硬い（伸縮しにくい）
  for(int i = 0; i < ropeBody->m_materials.size(); i++){
    ropeBody->m_materials[i]->m_kLST = 0.5; 
  }

  // 衝突判定に余裕を持たせるための隙間の幅（法線方向）
  //   デフォルト値(0.25)のままだと，ロープと回転体の間の隙間が広く見える
  ropeBody->getCollisionShape()->setMargin(0.02);

  // 位置ソルバ反復計算回数 (positions solver iterations)．
  //   節点個数の多さ，etc. に応じて増やさないとロープの位置が安定しない
  //   デフォルト値(1)のままで節点個数が20程度以上だと安定しない
  ropeBody->m_cfg.piterations = 6; 
  
  
  // カメラ位置・姿勢の設定
  bpWorld.setCamera(0, 12, 3.5, // 注視点の位置 [x, y, z]
                    15, // 注視点からカメラ視点までの距離
                    0, // 方位角（azimuth） [deg]
                    15, // 仰角（elevation） [deg]
                    0, 1, 0); // 天井方向ベクトル [x, y, z] （cf. OpenGLのgluLookAt()）
}


////////////////////////////////////////////////////////////////////////////////
/**
 *  使用法を表示する
 */
void showUsage(){
  using namespace std;
  cout << "--- Usage ----------------------------------------\n"
       << " Press 'j'/'k' to increase/decrease the velocity\n"
       << " Press '0' to stop the rotor\n"
       << " Press 'i' to start/pause the simulation\n"
       << " Press Space to reset the scene\n"
       << "--------------------------------------------------" << endl;
}

////////////////////////////////////////////////////////////////////////////////
/**
 *  プログラムの開始点
 */
int main(){
  // 動力学シミュレーションのワールドの準備
  BPWorld bpWorld("Winch",   // ウィンドウのタイトル
                  640, 480,     // ウィンドウの幅と高さ [pixels]
                  BPWorld::SOFT_BODY_INCLUDED,
                  doMonitoring, // モニタリング用関数へのポインタ
                  checkKey);    // キーボード入力処理用関数へのポインタ
  
  setUpObjects(bpWorld); // 物体を作成して配置する
  showUsage();
  bpWorld.start(); // ウィンドウを表示して，シミュレーションを開始する

  return 0;
}
