////////////////////////////////////////////////////////////////////////////////
/**
 *  例題プログラム TriMesh:
 *     ファイルから読み込んだ三角メッシュの形状の物体（剛体，柔軟体）を作成する
 *
 *  ● 概要
 *    ファイルから読み込んだ同じ形状の剛体と柔軟体を作り，地面に落下させます
 *
 *  ● 読み込む三角メッシュのファイル形式: Wavefront OBJ (.obj)
 *    Wavefront OBJは旧 Wavefront Technologies 社が設計したファイル形式です。
 *    MayaやLightwaveなど多くの3D CGソフトウェアがWavefront OBJに対応しています。
 *    単に"OBJ"や"obj", "object"と表示されている場合もあります。
 *
 *    Wavefront OBJファイルの読み込みには次の cd_wavefront.h を使用します。
 *    <bullet-2.7x>\Extras\ConvexDecomposition\cd_wavefront.h
 *      このフォルダへのパスは プロジェクトファイルTrimesh.cbpに記述されていますので
 *      ヘッダファイル"cd_wavefront.h"をincludeできます。
 *
 *    これはデモ ConvexDecompositionDemo で使われています。
 *    このデモのソースファイルは次のとおりです。
 *      c:\bullet-2.75\Demos\ConvexDecompositionDemo\ConvexDecompositionDemo.cpp
 *
 *    なお <bullet-2.7x>\Extras 以下のフォルダはBPの配布ファイルに含まれていますが，
 *    "Extras"とあるようにBPとは別のプロジェクトで作られたものです。
 *    <bullet-2.7x>\Extras\ConvexDecomposition以下のファイルの作者は
 *    NVIDIA社のJohn Ratcliff氏です。
 *     John Ratcliff氏によるWebページ（ConvexDecomposition他を入手可能）:
 *       - http://code.google.com/p/codesuppository/source/checkout
 *       - http://www.amillionpixels.us/sourcecode.htm
 *
 *
 *  ● 使用するOBJファイル
 *    例として Utah Teapot ( http://ja.wikipedia.org/wiki/Utah_teapot ) を準備しました。
 *    これはGLUTの代替となるオープンソースライブラリの freeglut
 *    （ http://freeglut.sourceforge.net/ ） のソースファイルに含まれている
 *    Utah Teapot の三角メッシュデータから作成したものです。
 *    元データは C言語の配列としてソースファイル freeglut_teapot_data.h にあります。
 *    利用した freeglut のバージョンは 2.6.0 です。
 *   
 *    freeglut は X-Consortiumライセンスで配布されています。詳細な著作権表示については，
 *    下記OBJファイルのコメント部分をご覧ください（英文）。
 *   
 *   ファイル（本ソースファイルと同じフォルダにあり）
 *     - teapot_with_cut.obj:
 *        freeglut の Utah Teapot の三角メッシュを縮小したものです。
 *        縮小には Metasequoia LE R2.3b を使用しました。
 *        このままだと注ぎ口と取っ手の部分が本体と別の塊になっています。
 *   
 *     - teapot_closed.obj:
 *       注ぎ口と取っ手の部分が本体と一体となるように上記三角メッシュに面を追加したものです。
 *       面の追加には Metasequoia LE R2.3b を使用しました。
 *       なお Metasequoia LE R2.3b は直接OBJを出力できないので次のように変換しました:
 *          Metasequoiaから.xファイルを出力 → Blender 2.49bで取り込んで，OBJファイルを出力
 *   
 *        Metasequoia 公式 Webサイト:  http://www.metaseq.net/metaseq/
 *        Blender 公式 Webサイト:      http://www.blender.org/
 *
 *  ● btBoxShape等の基本形状の物体と異なり，三角メッシュの形状(btGImpactMeshShape)の物体
 *     の衝突判定には以下の呼び出しが必要です。
 *        btGImpactCollisionAlgorithm::registerAlgorithm(m_dispatcher);
 *
 *       下記API Reference の項目"Detailed Description"を参照
 *         http://bulletphysics.com/Bullet/BulletFull/classbtGImpactCollisionAlgorithm.html
 *
 *    上記の行はBPWorld.cppのinitPhysics()に記述されています。
 *
 *     btGImpactMeshShapeについては "Bullet User Manual", p.16, p.19 に言及があります。
 *
 *     公式Webページトップからの辿り方
 *       http://bulletphysics.org/
 *       -> ページ左上端のリンク"Wiki"
 *       -> リンク"Bullet User Manual and API documentation"
 *       -> 項目"Bullet User Manual"以下のリンク"PDF version"
 *        (PDF中のタイトルは"Bullet 2.76 Physics SDK Manual")
 *
 *     なお名称GIMPACTは，現在BPに取り込まれていますが元々は別の衝突検出ライブラリ
 *     GIMPACTに由来します。
 *       GIMPACT開発者 Francisco Leon 氏のWebページ: http://gimpact.sourceforge.net/
 *
 *  ● 参考にしたデモ:
 *    - "SoftBody Cluster Car"
 *     （C:\bullet-2.7x\Demos\AllBulletDemos\AppAllBulletDemos.exe のボックスTestsで選択し実行）
 *      ソースファイル: C:\bullet-2.75\Demos\SoftDemo\SoftDemo.cpp内， 関数Init_ClusterCar()
 *    - "GIMPACT Test Demo"
 *      c:\bullet-2.7x\Demos\GimpactTestDemo
 *
 */

#include "BPWorld.h"
#include <BulletCollision/Gimpact/btGImpactShape.h>
#include <iostream>
#include <cstdlib>
#include "cd_wavefront.h" // ConvexDecomposition::WavefrontObjを使うため

// 下記二つはプログラム実行中に頂点データが解放しないようにするため，
// グローバル変数にする
ConvexDecomposition::WavefrontObj g_objData; // OBJファイルを読み込むクラス
btTriangleIndexVertexArray *g_idxVtxArray; // BPで三角メッシュを扱うためのクラス。
// --> BPの設計意図としては特定のファイル形式に依存しないようにファイルを
//     読み込むクラスとbtTriangleIndexVertexArrayの 2段構えにしている


////////////////////////////////////////////////////////////////////////////////
/**
 *  物体を作成して，ワールドに配置する
 */
void setUpObjects(BPWorld& bpWorld){
  using namespace std;
  const char *objFileName = "teapot_closed.obj"; // 閉じたティーポット
  // const char *objFileName = "teapot_with_cut.obj"; // 一部が途切れたティーポット

  unsigned int result = g_objData.loadObj(objFileName); // OBJファイルを読み込む
  if(result == 0){
    cout << "Error: Can't find " << objFileName << endl;
    exit(EXIT_FAILURE);
  }

  g_idxVtxArray = new btTriangleIndexVertexArray // BPが三角メッシュを扱うためのクラス
    (g_objData.mTriCount, // 三角形の個数
     g_objData.mIndices,  // 三角形を表す三つ組頂点インデックスの配列の先頭へのポインタ
     3*sizeof(g_objData.mIndices[0]), // 三つ組頂点インデックス(v1,v2,v3) のサイズ [bytes]
     g_objData.mVertexCount, // 頂点の個数
     g_objData.mVertices, // 頂点データの配列の先頭へのポインタ
     3*sizeof(g_objData.mVertices[0])); // 頂点データ [x, y, z] 1個分のサイズ [bytes]

  // 三角メッシュの形状を表すクラス
  btGImpactMeshShape *triMeshShape = new btGImpactMeshShape(g_idxVtxArray);

  // 以下は衝突判定のために必要な呼び出し。
  // 下記API Referenceの項目"Detailed Description"を参照
  //   http://bulletphysics.com/Bullet/BulletFull/classbtGImpactMeshShape.html
  triMeshShape->updateBound();


  btTransform transform;
  transform.setIdentity();
  transform.setOrigin(btVector3(-2, 2, 0)); // 剛体の初期位置

  // 剛体を作る
  bpWorld.createRigidBody(1.0,        // 質量
                          transform,  // 初期位置，姿勢
                          triMeshShape); // 衝突形状

  // 柔軟体を作る
  cout << "Constructing a soft body..." << endl;
  btSoftBody *triMeshSoftBody = btSoftBodyHelpers::CreateFromTriMesh
    (bpWorld.getSoftBodyWorldInfo(),
     g_objData.mVertices, // 頂点データ配列の先頭へのポインタ（型btScalar*）
                          // （btScalar≠floatのときは 実行時エラーとなるので注意。
                          //   デフォルトではbtScalar = float）
     g_objData.mIndices, // 三つ組頂点インデックスの配列の先頭へのポインタ
     g_objData.mTriCount); // 三角形の個数

  transform.setOrigin(btVector3(2, 2, 0)); // 初期位置
  triMeshSoftBody->transform(transform); // 初期位置・姿勢を設定する

  triMeshSoftBody->setTotalMass(1.0, // 質量
                                true);// trueなら各三角面の面積を基に質量の分布を定める
                                      // falseなら現在の質量の比分布を維持して
                                      // 全体の質量が第一引数の値になるようにする

  btSoftBody::Material* pm = triMeshSoftBody->appendMaterial(); // リンクの材質情報を追加
  pm->m_flags -= btSoftBody::fMaterial::DebugDraw; // 処理を速くするため追加分のリンクを非表示に設定
  pm->m_kLST = 0.6; // 追加分のリンクの剛性（堅さ）。値の範囲[0,1]。 大きいほど潰れにくい

  cout << "Configuring the soft body... (This will take several seconds)" << endl;

  // 柔軟体が潰れないようにリンクを追加する
  //  第1引数 distance:
  //    新たにリンクで接続するノード間の現在の距離（この距離にあるノード2個を探索して接続する）
  //    ただしここで 距離 = 辿るのに経由するリンクの数（cf. グラフ理論）。
  //    距離 2とすると 四角を作っている4個のノードの対角のノードを接続するリンクを作る。
  //    （ただし元から三角メッシュなので重複するリンクは作らない）
  //
  //    ある4個のノード     リンクを追加した状態
  //     Ａ─Ｂ　　　　　　　Ａ─Ｂ
  //     ｜／｜  　　　→    ｜Ｘ｜
  //     Ｃ─Ｄ　　　　　　　Ｃ─Ｄ
  //
  //     ノード{A, D}間の距離が2 
  //
  //  第2引数 material: 作成するリンクの材質
  triMeshSoftBody->generateBendingConstraints(2,   // 距離
                                              pm); // 材質

  triMeshSoftBody->m_cfg.piterations = 4; // 繰り返し計算の回数の設定（多いほうが正確）

  // 小さくしないと地面に落ちた後に浮いて見えるので以下のmarginを調整する
  // （例題Winchでも同様の処理をしておりテキスト中に解説あり）
  triMeshSoftBody->getCollisionShape()->setMargin(0.01);


  // 衝突判定に関して衝突形状を複数の凸包の塊(cluster, クラスタ)に自動分割するように設定。
  // （デフォルトでは各節点と各面をそのまま衝突判定に使う）
  //  参照:
  //    ● デモ "SoftBody Cluster Car"（C:\bullet-2.7x\Demos\AllBulletDemos\AppAllBulletDemos.exe）
  //       ソースファイル C:\bullet-2.75\Demos\SoftDemo\SoftDemo.cpp内， 関数Init_ClusterCar()
  //    ● "Bullet User Manual", p.30, 項目"Collision clusters"
  triMeshSoftBody->generateClusters(3); // 引数: クラスタの個数

  // 上記のクラスタを使った衝突判定を有効にする
  triMeshSoftBody->m_cfg.collisions =
    btSoftBody::fCollision::CL_RS // "CLuster Rigid vs Soft"の略
    | btSoftBody::fCollision::CL_SS;  // "CLuster Soft vs Soft"の略


  bpWorld.addSoftBody(triMeshSoftBody); // 柔軟体をBPワールドに追加する

  // カメラの位置・姿勢の設定
  bpWorld.setCamera(0, 0, 0, // 注視点の位置 [x, y, z]
                    4,  // 注視点からカメラ視点までの距離
                    0,  // 方位角（azimuth） [deg]
                    30, // 仰角（elevation） [deg]
                    0, 1, 0); // 天井方向ベクトル [x, y, z] （cf. OpenGLのgluLookAt()）
}


////////////////////////////////////////////////////////////////////////////////
/**
 *  プログラムの開始点
 */
int main(){
  // 動力学シミュレーションのワールドの準備
  BPWorld bpWorld("TriMesh", // ウィンドウのタイトル
                  640, 480,  // ウィンドウの幅と高さ [pixels]
                  BPWorld::SOFT_BODY_INCLUDED);
  setUpObjects(bpWorld); // 物体を作成して配置する
  bpWorld.setGlobalERP(0.2);
  bpWorld.toggleIdle();
  std::cout << "Press 'i' to start" << std::endl;
  bpWorld.start(); // ウィンドウを表示して，シミュレーションを開始する

  delete g_idxVtxArray;
  return 0;
}
